# Copyright (c) 2012 Jorge Hernandez - Melendez

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# TODO : prop names into English, add missing tooltips

bl_info = {
    "name": "Rope Creator",
    "description": "Dynamic rope (with cloth) creator",
    "author": "Jorge Hernandez - Melenedez",
    "version": (0, 2, 2),
    "blender": (2, 7, 3),
    "location": "Left Toolbar > ClothRope",
    "warning": "",
    "wiki_url": "",
    "category": "Add Mesh"
}


import bpy
from bpy.types import Operator
from bpy.props import (
        BoolProperty,
        FloatProperty,
        IntProperty,
        )


def desocultar(quien):
    if quien == "todo":
        for ob in bpy.data.objects:
            ob.hide = False
    else:
        bpy.data.objects[quien].hide = False


def deseleccionar_todo():
    bpy.ops.object.select_all(action='DESELECT')


def seleccionar_todo():
    bpy.ops.object.select_all(action='SELECT')


def salir_de_editmode():
    if bpy.context.mode in ["EDIT", "EDIT_MESH", "EDIT_CURVE"]:
        bpy.ops.object.mode_set(mode='OBJECT')


# Clear scene:
def reset_scene():
    desocultar("todo")
    # playback to the start
    bpy.ops.screen.frame_jump(end=False)
    try:
        salir_de_editmode()
    except:
        pass
    try:
        area = bpy.context.area
        # expand everything in the outliner to be able to select children
        old_type = area.type
        area.type = 'OUTLINER'
        bpy.ops.outliner.expanded_toggle()

        # restore the original context
        area.type = old_type

        seleccionar_todo()
        bpy.ops.object.delete(use_global=False)

    except Exception as e:
        print("\n[rope_alpha]\nfunction: reset_scene\nError: %s" % e)


def entrar_en_editmode():
    if bpy.context.mode == "OBJECT":
        bpy.ops.object.mode_set(mode='EDIT')


def select_all_in_edit_mode(ob):
    if ob.mode != 'EDIT':
        entrar_en_editmode()
    bpy.ops.mesh.select_all(action="DESELECT")
    bpy.context.tool_settings.mesh_select_mode = (True, False, False)
    salir_de_editmode()
    for v in ob.data.vertices:
        if not v.select:
            v.select = True
    entrar_en_editmode()


def deselect_all_in_edit_mode(ob):
    if ob.mode != 'EDIT':
        entrar_en_editmode()
    bpy.ops.mesh.select_all(action="DESELECT")
    bpy.context.tool_settings.mesh_select_mode = (True, False, False)
    salir_de_editmode()
    for v in ob.data.vertices:
        if not v.select:
            v.select = False
    entrar_en_editmode()


def which_vertex_are_selected(ob):
    for v in ob.data.vertices:
        if v.select:
            print(str(v.index))
            print("Vertex " + str(v.index) + " is selected")


def seleccionar_por_nombre(nombre):
    scn = bpy.context.scene
    bpy.data.objects[nombre].select = True

    scn.objects.active = bpy.data.objects[nombre]


def deseleccionar_por_nombre(nombre):
    bpy.data.objects[nombre].select = False


def crear_vertices(ob):
    ob.data.vertices.add(1)
    ob.data.update


def borrar_elementos_seleccionados(tipo):
    if tipo == "vertices":
        bpy.ops.mesh.delete(type='VERT')


def obtener_coords_vertex_seleccionados():
    coordenadas_de_vertices = []
    for ob in bpy.context.selected_objects:
        if ob.type == 'MESH':
            for v in ob.data.vertices:
                if v.select:
                    coordenadas_de_vertices.append([v.co[0], v.co[1], v.co[2]])
            return coordenadas_de_vertices[0]


def crear_locator(pos):
    bpy.ops.object.empty_add(
            type='PLAIN_AXES', radius=1, view_align=False,
            location=(pos[0], pos[1], pos[2]),
            layers=(True, False, False, False, False, False, False,
                    False, False, False, False, False, False, False,
                    False, False, False, False, False, False)
            )


def extruir_vertices(longitud, cuantos_segmentos):
    bpy.ops.mesh.extrude_region_move(
            MESH_OT_extrude_region={"mirror": False},
            TRANSFORM_OT_translate={
                    "value": (longitud / cuantos_segmentos, 0, 0),
                    "constraint_axis": (True, False, False),
                    "constraint_orientation": 'GLOBAL', "mirror": False,
                    "proportional": 'DISABLED', "proportional_edit_falloff": 'SMOOTH',
                    "proportional_size": 1, "snap": False, "snap_target": 'CLOSEST',
                    "snap_point": (0, 0, 0), "snap_align": False, "snap_normal": (0, 0, 0),
                    "gpencil_strokes": False, "texture_space": False,
                    "remove_on_cancel": False, "release_confirm": False
                    }
            )


def select_all_vertex_in_curve_bezier(bc):
    for i in range(len(bc.data.splines[0].points)):
        bc.data.splines[0].points[i].select = True


def deselect_all_vertex_in_curve_bezier(bc):
    for i in range(len(bc.data.splines[0].points)):
        bc.data.splines[0].points[i].select = False


def ocultar_relationships():
    for area in bpy.context.screen.areas:
        if area.type == 'VIEW_3D':
            area.spaces[0].show_relationship_lines = False


class ClothRope(Operator):
    bl_idname = "clot.rope"
    bl_label = "Rope Cloth"
    bl_description = ("Create a new Scene with a Cloth modifier\n"
                      "Rope Simulation with hooked Helper Objects")

    ropelenght = IntProperty(
            name="Rope Length",
            description="Length of the generated Rope",
            default=5
            )
    ropesegments = IntProperty(
            name="Rope Segments",
            description="Number of the Rope Segments",
            default=5
            )
    qcr = IntProperty(
            name="Collision Quality",
            description="Rope's Cloth modifier collsion quality",
            min=1, max=20,
            default=20
            )
    substeps = IntProperty(
            name="Rope Substeps",
            description="Rope's Cloth modifier quality",
            min=4, max=80,
            default=50
            )
    resrope = IntProperty(
            name="Rope Resolution",
            description="Rope's Bevel resolution",
            default=5
            )
    radiusrope = FloatProperty(
            name="Radius",
            description="Rope's Radius",
            min=0.04, max=1,
            default=0.04
            )
    hide_emptys = BoolProperty(
            name="Hide Empties",
            description="Hide Helper Objects",
            default=False
            )

    def execute(self, context):
        # add a new scene
        bpy.ops.scene.new(type="NEW")
        scene = bpy.context.scene
        scene.name = "Test Rope"
        seleccionar_todo()
        longitud = self.ropelenght

        # For the middle to have x segments between the first and
        # last point, must add 1 to the quantity:
        cuantos_segmentos = self.ropesegments + 1
        calidad_de_colision = self.qcr
        substeps = self.substeps
        deseleccionar_todo()
        # collect the possible empties that already exist in the data
        empties_prev = [obj.name for obj in bpy.data.objects if obj.type == "EMPTY"]

        # create an empty that will be the parent of everything
        bpy.ops.object.empty_add(
                type='SPHERE', radius=1, view_align=False, location=(0, 0, 0),
                layers=(True, False, False, False, False, False, False, False,
                        False, False, False, False, False, False, False, False,
                        False, False, False, False)
                )
        ob = bpy.context.selected_objects[0]
        ob.name = "Rope"
        # .001 and friends
        rope_name = ob.name
        deseleccionar_todo()

        # create a plane and delete it
        bpy.ops.mesh.primitive_plane_add(
                radius=1, view_align=False, enter_editmode=False, location=(0, 0, 0),
                layers=(True, False, False, False, False, False, False, False, False,
                        False, False, False, False, False, False, False, False,
                        False, False, False)
                )
        ob = bpy.context.selected_objects[0]
        # rename:
        ob.name = "cuerda"
        # .001 and friends
        cuerda_1_name = ob.name

        entrar_en_editmode()  # enter edit mode
        select_all_in_edit_mode(ob)

        borrar_elementos_seleccionados("vertices")
        salir_de_editmode()  # leave edit mode
        crear_vertices(ob)  # create a vertex

        # Creating a Group for the PIN
        # Group contains the vertices of the pin and the Group.001 contains the single main line
        entrar_en_editmode()  # enter edit mode
        bpy.ops.object.vertex_group_add()  # create a group
        select_all_in_edit_mode(ob)
        bpy.ops.object.vertex_group_assign()  # assign it

        salir_de_editmode()  # leave edit mode
        ob.vertex_groups[0].name = "Pin"
        deseleccionar_todo()
        seleccionar_por_nombre(cuerda_1_name)

        # extrude vertices:
        for i in range(cuantos_segmentos):
            entrar_en_editmode()
            extruir_vertices(longitud, cuantos_segmentos)
            # delete the PIN group
            bpy.ops.object.vertex_group_remove_from()
            # get the direction to create the locator on it's position
            pos = obtener_coords_vertex_seleccionados()

            salir_de_editmode()  # leave edit mode
            # create locator at position
            crear_locator(pos)
            deseleccionar_todo()
            seleccionar_por_nombre(cuerda_1_name)
        deseleccionar_todo()

        seleccionar_por_nombre(cuerda_1_name)  # select the rope
        entrar_en_editmode()

        pos = obtener_coords_vertex_seleccionados()  # get their positions
        salir_de_editmode()
        # create the last locator
        crear_locator(pos)
        deseleccionar_todo()
        seleccionar_por_nombre(cuerda_1_name)
        entrar_en_editmode()  # enter edit mode
        bpy.ops.object.vertex_group_add()  # Creating Master guide group
        select_all_in_edit_mode(ob)
        bpy.ops.object.vertex_group_assign()  # and assing it
        ob.vertex_groups[1].name = "Guide_rope"

        # extrude the Curve so it has a minumum thickness for collide
        bpy.ops.mesh.extrude_region_move(
                MESH_OT_extrude_region={"mirror": False},
                TRANSFORM_OT_translate={
                        "value": (0, 0.005, 0), "constraint_axis": (False, True, False),
                        "constraint_orientation": 'GLOBAL', "mirror": False,
                        "proportional": 'DISABLED', "proportional_edit_falloff": 'SMOOTH',
                        "proportional_size": 1, "snap": False, "snap_target": 'CLOSEST',
                        "snap_point": (0, 0, 0), "snap_align": False, "snap_normal": (0, 0, 0),
                        "gpencil_strokes": False, "texture_space": False,
                        "remove_on_cancel": False, "release_confirm": False
                        }
                )
        bpy.ops.object.vertex_group_remove_from()
        deselect_all_in_edit_mode(ob)
        salir_de_editmode()
        bpy.ops.object.modifier_add(type='CLOTH')
        bpy.context.object.modifiers["Cloth"].settings.use_pin_cloth = True
        bpy.context.object.modifiers["Cloth"].settings.vertex_group_mass = "Pin"
        bpy.context.object.modifiers["Cloth"].collision_settings.collision_quality = calidad_de_colision
        bpy.context.object.modifiers["Cloth"].settings.quality = substeps

        # Duplicate to convert into Curve:
        # select the vertices that are the part of the Group.001
        seleccionar_por_nombre(cuerda_1_name)
        entrar_en_editmode()
        bpy.ops.mesh.select_all(action="DESELECT")
        bpy.context.tool_settings.mesh_select_mode = (True, False, False)
        salir_de_editmode()
        gi = ob.vertex_groups["Guide_rope"].index  # get group index

        for v in ob.data.vertices:
            for g in v.groups:
                if g.group == gi:  # compare with index in VertexGroupElement
                    v.select = True

        # now we have to make a table of names of cuerdas to see which one will be new
        cuerda_names = [obj.name for obj in bpy.data.objects if "cuerda" in obj.name]

        entrar_en_editmode()

        # we already have the selected guide:
        # duplicate it:
        bpy.ops.mesh.duplicate_move(
                MESH_OT_duplicate={"mode": 1},
                TRANSFORM_OT_translate={
                        "value": (0, 0, 0), "constraint_axis": (False, False, False),
                        "constraint_orientation": 'GLOBAL', "mirror": False,
                        "proportional": 'DISABLED', "proportional_edit_falloff": 'SMOOTH',
                        "proportional_size": 1, "snap": False, "snap_target": 'CLOSEST',
                        "snap_point": (0, 0, 0), "snap_align": False, "snap_normal": (0, 0, 0),
                        "gpencil_strokes": False, "texture_space": False,
                        "remove_on_cancel": False, "release_confirm": False
                        }
                )
        # separate the selections:
        bpy.ops.mesh.separate(type='SELECTED')
        salir_de_editmode()
        deseleccionar_todo()

        cuerda_2_name = "cuerda.001"
        test = []
        for obj in bpy.data.objects:
            if "cuerda" in obj.name and obj.name not in cuerda_names:
                cuerda_2_name = obj.name
                test.append(obj.name)

        seleccionar_por_nombre(cuerda_2_name)

        # from the newly created curve remove the Cloth:
        bpy.ops.object.modifier_remove(modifier="Cloth")
        # convert the Curve:
        bpy.ops.object.convert(target='CURVE')

        # all Empties that are not previously present
        emptys = []
        for eo in bpy.data.objects:
            if eo.type == 'EMPTY' and eo.name not in empties_prev:
                if eo.name != rope_name:
                    emptys.append(eo)

        # select and deselect:
        bc = bpy.data.objects[cuerda_2_name]
        n = 0

        for e in emptys:
            deseleccionar_todo()
            seleccionar_por_nombre(e.name)
            seleccionar_por_nombre(bc.name)
            entrar_en_editmode()
            deselect_all_vertex_in_curve_bezier(bc)
            bc.data.splines[0].points[n].select = True
            bpy.ops.object.hook_add_selob(use_bone=False)
            salir_de_editmode()
            n = n + 1

        ob = bpy.data.objects[cuerda_1_name]
        n = 0

        for e in emptys:
            deseleccionar_todo()
            seleccionar_por_nombre(e.name)
            seleccionar_por_nombre(ob.name)
            entrar_en_editmode()
            bpy.ops.mesh.select_all(action="DESELECT")
            bpy.context.tool_settings.mesh_select_mode = (True, False, False)
            salir_de_editmode()

            for v in ob.data.vertices:
                if v.select:
                    v.select = False
            ob.data.vertices[n].select = True
            entrar_en_editmode()
            bpy.ops.object.vertex_parent_set()

            salir_de_editmode()
            n = n + 1

            # hide the Empties:
            deseleccionar_todo()

        # all parented to the spherical empty:
        seleccionar_por_nombre(cuerda_2_name)
        seleccionar_por_nombre(cuerda_1_name)
        seleccionar_por_nombre(rope_name)
        bpy.ops.object.parent_set(type='OBJECT', keep_transform=True)
        deseleccionar_todo()

        # do not display the relations
        ocultar_relationships()
        seleccionar_por_nombre(cuerda_2_name)

        # curved rope settings:
        bpy.context.object.data.fill_mode = 'FULL'
        bpy.context.object.data.bevel_depth = self.radiusrope
        bpy.context.object.data.bevel_resolution = self.resrope

        return {'FINISHED'}

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self, width=350)

    def draw(self, context):
        layout = self.layout
        box = layout.box()
        col = box.column(align=True)

        col.label("Rope settings:")
        rowsub0 = col.row()
        rowsub0.prop(self, "ropelenght", text="Length")
        rowsub0.prop(self, "ropesegments", text="Segments")
        rowsub0.prop(self, "radiusrope", text="Radius")

        col.label("Quality Settings:")
        col.prop(self, "resrope", text="Resolution curve")
        col.prop(self, "qcr", text="Quality Collision")
        col.prop(self, "substeps", text="Substeps")


class BallRope(Operator):
    bl_idname = "ball.rope"
    bl_label = "Wrecking Ball"
    bl_description = ("Create a new Scene with a Rigid Body simulation of\n"
                      "Wrecking Ball on a rope")

    # defaults rope ball
    ropelenght2 = IntProperty(
            name="Rope Length",
            description="Length of the Wrecking Ball rope",
            default=10
            )
    ropesegments2 = IntProperty(
            name="Rope Segments",
            description="Number of the Wrecking Ball rope segments",
            min=0, max=999,
            default=6
            )
    radiuscubes = FloatProperty(
            name="Cube Radius",
            description="Size of the Linked Cubes helpers",
            default=0.5
            )
    radiusrope = FloatProperty(
            name="Rope Radius",
            description="Radius of the Rope",
            default=0.4
            )
    worldsteps = IntProperty(
            name="World Steps",
            description="Rigid Body Solver world steps per second (update)",
            min=60, max=1000,
            default=250
            )
    solveriterations = IntProperty(
            name="Solver Iterations",
            description="How many times the Rigid Body Solver should run",
            min=10, max=100,
            default=50
            )
    massball = IntProperty(
            name="Ball Mass",
            description="Mass of the Wrecking Ball",
            default=1
            )
    resrope = IntProperty(
            name="Resolution",
            description="Rope resolution",
            default=4
            )
    grados = FloatProperty(
            name="Degrees",
            description="Angle of the Wrecking Ball compared to the Ground Plane",
            default=45
            )
    separacion = FloatProperty(
            name="Link Cubes Gap",
            description="Space between the Rope's Linked Cubes",
            default=0.1
            )
    hidecubes = BoolProperty(
            name="Hide Link Cubes",
            description="Hide helper geometry for the Rope",
            default=False
            )

    def execute(self, context):
        world_steps = self.worldsteps
        solver_iterations = self.solveriterations
        longitud = self.ropelenght2

        # make a + 2, so the segments will be between the two end points...
        segmentos = self.ropesegments2 + 2
        offset_del_suelo = 1
        offset_del_suelo_real = (longitud / 2) + (segmentos / 2)
        radio = self.radiuscubes
        radiorope = self.radiusrope
        masa = self.massball
        resolucion = self.resrope
        rotrope = self.grados
        separation = self.separacion
        hidecubeslinks = self.hidecubes

        # add new scene
        bpy.ops.scene.new(type="NEW")
        scene = bpy.context.scene
        scene.name = "Test Ball"

        # collect the possible constraint empties that already exist in the data
        constraint_prev = [obj.name for obj in bpy.data.objects if
                           obj.type == "EMPTY" and "Constraint" in obj.name]
        # floor:
        bpy.ops.mesh.primitive_cube_add(
                radius=1, view_align=False, enter_editmode=False, location=(0, 0, 0),
                layers=(True, False, False, False, False, False, False, False, False,
                        False, False, False, False, False, False, False, False,
                        False, False, False)
                )
        bpy.context.object.scale.x = 10 + longitud
        bpy.context.object.scale.y = 10 + longitud
        bpy.context.object.scale.z = 0.05
        bpy.context.object.name = "groundplane"
        # The secret agents .001, 002 etc.
        groundplane_name = bpy.context.object.name

        bpy.ops.rigidbody.objects_add(type='PASSIVE')

        # create the first cube:
        cuboslink = []
        n = 0
        for i in range(segmentos):
            # if 0 start from 1
            if i == 0:
                i = offset_del_suelo
            else:  # if it is not 0, add one so it doesn't step on the first one starting from 1
                i = i + offset_del_suelo
            separacion = longitud * 2 / segmentos  # distance between linked cubes
            bpy.ops.mesh.primitive_cube_add(
                    radius=1, view_align=False, enter_editmode=False,
                    location=(0, 0, i * separacion),
                    layers=(True, False, False, False, False, False, False, False,
                            False, False, False, False, False, False, False, False,
                            False, False, False, False)
                    )
            bpy.ops.rigidbody.objects_add(type='ACTIVE')
            bpy.context.object.name = "CubeLink"
            if n != 0:
                bpy.context.object.draw_type = 'WIRE'
                bpy.context.object.hide_render = True
            n += 1
            bpy.context.object.scale.z = (longitud * 2) / (segmentos * 2) - separation
            bpy.context.object.scale.x = radio
            bpy.context.object.scale.y = radio
            cuboslink.append(bpy.context.object)

        for i in range(len(cuboslink)):
            deseleccionar_todo()
            if i != len(cuboslink) - 1:
                nombre1 = cuboslink[i]
                nombre2 = cuboslink[i + 1]
                seleccionar_por_nombre(nombre1.name)
                seleccionar_por_nombre(nombre2.name)
                bpy.ops.rigidbody.connect()

        # select by name
        constraint_new = [
                    obj.name for obj in bpy.data.objects if
                    obj.type == "EMPTY" and "Constraint" in obj.name and
                    obj.name not in constraint_prev
                    ]

        for names in constraint_new:
            seleccionar_por_nombre(names)

        for c in bpy.context.selected_objects:
            c.rigid_body_constraint.type = 'POINT'
        deseleccionar_todo()

        # create a Bezier curve:
        bpy.ops.curve.primitive_bezier_curve_add(
                radius=1, view_align=False, enter_editmode=False, location=(0, 0, 0),
                layers=(True, False, False, False, False, False, False, False, False,
                False, False, False, False, False, False, False, False, False, False, False)
                )
        bpy.context.object.name = "Cuerda"
        # Blender will automatically append the .001
        # if it is already in data
        real_name = bpy.context.object.name

        for i in range(len(cuboslink)):
            cubonombre = cuboslink[i].name
            seleccionar_por_nombre(cubonombre)
            seleccionar_por_nombre(real_name)
            x = cuboslink[i].location[0]
            y = cuboslink[i].location[1]
            z = cuboslink[i].location[2]

            # if it is 0 make it start from 1 as the offset from the ground...
            if i == 0:
                i = offset_del_suelo
            else:  # if it is not 0, add one so it doesn't step on the first one starting from 1
                i = i + offset_del_suelo

            salir_de_editmode()
            entrar_en_editmode()

            if i == 1:
                # select all the vertices and delete them
                select_all_vertex_in_curve_bezier(bpy.data.objects[real_name])
                bpy.ops.curve.delete(type='VERT')
                # create the first vertex:
                bpy.ops.curve.vertex_add(location=(x, y, z))
            else:
                # extrude the rest:
                bpy.ops.curve.extrude_move(
                        CURVE_OT_extrude={"mode": 'TRANSLATION'},
                        TRANSFORM_OT_translate={
                            "value": (0, 0, z / i),
                            "constraint_axis": (False, False, True),
                            "constraint_orientation": 'GLOBAL', "mirror": False,
                            "proportional": 'DISABLED', "proportional_edit_falloff": 'SMOOTH',
                            "proportional_size": 1, "snap": False, "snap_target": 'CLOSEST',
                            "snap_point": (0, 0, 0), "snap_align": False, "snap_normal": (0, 0, 0),
                            "gpencil_strokes": False, "texture_space": False,
                            "remove_on_cancel": False, "release_confirm": False
                            }
                        )
            bpy.ops.object.hook_add_selob(use_bone=False)
            salir_de_editmode()
            bpy.context.object.data.bevel_resolution = resolucion
            deseleccionar_todo()

        # create a sphere ball:
        deseleccionar_todo()
        seleccionar_por_nombre(cuboslink[0].name)
        entrar_en_editmode()
        z = cuboslink[0].scale.z + longitud / 2
        bpy.ops.view3d.snap_cursor_to_selected()
        bpy.ops.mesh.primitive_uv_sphere_add(
                view_align=False, enter_editmode=False,
                layers=(True, False, False, False, False, False, False,
                        False, False, False, False, False, False, False,
                        False, False, False, False, False, False)
                )
        bpy.ops.transform.translate(
                value=(0, 0, -z + 2), constraint_axis=(False, False, True),
                constraint_orientation='GLOBAL', mirror=False, proportional='DISABLED',
                proportional_edit_falloff='SMOOTH', proportional_size=1
                )
        bpy.ops.transform.resize(
                value=(longitud / 2, longitud / 2, longitud / 2),
                constraint_axis=(False, False, False),
                constraint_orientation='GLOBAL',
                mirror=False, proportional='DISABLED',
                proportional_edit_falloff='SMOOTH', proportional_size=1
                )
        deselect_all_in_edit_mode(cuboslink[0])
        salir_de_editmode()
        bpy.ops.object.shade_smooth()
        bpy.context.object.rigid_body.mass = masa
        bpy.ops.object.origin_set(type='ORIGIN_CENTER_OF_MASS')

        # move it all up a bit more:
        seleccionar_todo()
        deseleccionar_por_nombre(groundplane_name)
        bpy.ops.transform.translate(
                value=(0, 0, offset_del_suelo_real),
                constraint_axis=(False, False, True),
                constraint_orientation='GLOBAL', mirror=False,
                proportional='DISABLED', proportional_edit_falloff='SMOOTH',
                proportional_size=1
                )

        deseleccionar_todo()
        seleccionar_por_nombre(cuboslink[-1].name)
        bpy.ops.rigidbody.objects_add(type='PASSIVE')

        bpy.context.scene.rigidbody_world.steps_per_second = world_steps
        bpy.context.scene.rigidbody_world.solver_iterations = solver_iterations

        # move everything from the top one:
        seleccionar_por_nombre(cuboslink[-1].name)
        bpy.ops.view3d.snap_cursor_to_selected()
        seleccionar_todo()
        deseleccionar_por_nombre(groundplane_name)
        deseleccionar_por_nombre(cuboslink[-1].name)
        bpy.context.space_data.pivot_point = 'CURSOR'
        bpy.ops.transform.rotate(
                value=rotrope, axis=(1, 0, 0),
                constraint_axis=(True, False, False),
                constraint_orientation='GLOBAL',
                mirror=False, proportional='DISABLED',
                proportional_edit_falloff='SMOOTH',
                proportional_size=1
                )
        bpy.context.space_data.pivot_point = 'MEDIAN_POINT'
        deseleccionar_todo()

        seleccionar_por_nombre(real_name)
        bpy.context.object.data.fill_mode = 'FULL'
        bpy.context.object.data.bevel_depth = radiorope
        for ob in bpy.data.objects:
            if ob.name != cuboslink[0].name:
                if ob.name.find("CubeLink") >= 0:
                    deseleccionar_todo()
                    seleccionar_por_nombre(ob.name)
                    if hidecubeslinks:
                        bpy.context.object.hide = True
        ocultar_relationships()
        deseleccionar_todo()
        return {'FINISHED'}

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self, width=350)

    def draw(self, context):
        layout = self.layout
        box = layout.box()
        col = box.column(align=True)

        col.label("Rope settings:")
        rowsub0 = col.row()
        rowsub0.prop(self, "hidecubes", text="Hide Link Cubes")

        rowsub1 = col.row(align=True)
        rowsub1.prop(self, "ropelenght2", text="Length")
        rowsub1.prop(self, "ropesegments2", text="Segments")

        rowsub2 = col.row(align=True)
        rowsub2.prop(self, "radiuscubes", text="Radius Link Cubes")
        rowsub2.prop(self, "radiusrope", text="Radius Rope")

        rowsub3 = col.row(align=True)
        rowsub3.prop(self, "grados", text="Degrees")
        rowsub3.prop(self, "separacion", text="Separation Link Cubes")

        col.label("Quality Settings:")
        col.prop(self, "resrope", text="Resolution Rope")
        col.prop(self, "massball", text="Ball Mass")
        col.prop(self, "worldsteps", text="World Steps")
        col.prop(self, "solveriterations", text="Solver Iterarions")


# Register

def register():
    bpy.utils.register_module(__name__)


def unregister():
    bpy.utils.unregister_module(__name__)


if __name__ == "__main__":
    register()



import bpy

from .paw import Rig as pawRig
from .paw import parameters_ui
from .paw import add_parameters

IMPLEMENTATION = True   # Include and set True if Rig is just an implementation for a wrapper class
                        # add_parameters and parameters_ui are unused for implementation classes


class Rig(pawRig):

    def __init__(self, obj, bone_name, params):
        super(Rig, self).__init__(obj, bone_name, params)


def create_sample(obj):
    # generated by rigify.utils.write_metarig
    bpy.ops.object.mode_set(mode='EDIT')
    arm = obj.data

    bones = {}

    bone = arm.edit_bones.new('thigh.L')
    bone.head[:] = 0.0291, 0.1181, 0.2460
    bone.tail[:] = 0.0293, 0.1107, 0.1682
    bone.roll = 3.1383
    bone.use_connect = False
    bones['thigh.L'] = bone.name
    bone = arm.edit_bones.new('shin.L')
    bone.head[:] = 0.0293, 0.1107, 0.1682
    bone.tail[:] = 0.0293, 0.1684, 0.1073
    bone.roll = 3.1416
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['thigh.L']]
    bones['shin.L'] = bone.name
    bone = arm.edit_bones.new('foot.L')
    bone.head[:] = 0.0293, 0.1684, 0.1073
    bone.tail[:] = 0.0293, 0.1530, 0.0167
    bone.roll = 3.1416
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['shin.L']]
    bones['foot.L'] = bone.name
    bone = arm.edit_bones.new('r_toe.L')
    bone.head[:] = 0.0293, 0.1530, 0.0167
    bone.tail[:] = 0.0293, 0.1224, 0.0167
    bone.roll = 0.0000
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['foot.L']]
    bones['r_toe.L'] = bone.name
    bone = arm.edit_bones.new('r_palm.001.L')
    bone.head[:] = 0.0220, 0.1457, 0.0123
    bone.tail[:] = 0.0215, 0.1401, 0.0123
    bone.roll = 0.0014
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['r_toe.L']]
    bones['r_palm.001.L'] = bone.name
    bone = arm.edit_bones.new('r_palm.002.L')
    bone.head[:] = 0.0297, 0.1458, 0.0123
    bone.tail[:] = 0.0311, 0.1393, 0.0123
    bone.roll = -0.0005
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['r_toe.L']]
    bones['r_palm.002.L'] = bone.name
    bone = arm.edit_bones.new('r_palm.003.L')
    bone.head[:] = 0.0363, 0.1473, 0.0123
    bone.tail[:] = 0.0376, 0.1407, 0.0123
    bone.roll = 0.0000
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['r_toe.L']]
    bones['r_palm.003.L'] = bone.name
    bone = arm.edit_bones.new('r_palm.004.L')
    bone.head[:] = 0.0449, 0.1501, 0.0123
    bone.tail[:] = 0.0466, 0.1479, 0.0123
    bone.roll = -0.0004
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['r_toe.L']]
    bones['r_palm.004.L'] = bone.name
    bone = arm.edit_bones.new('r_index.001.L')
    bone.head[:] = 0.0215, 0.1367, 0.0087
    bone.tail[:] = 0.0217, 0.1325, 0.0070
    bone.roll = -0.3427
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['r_palm.001.L']]
    bones['r_index.001.L'] = bone.name
    bone = arm.edit_bones.new('r_middle.001.L')
    bone.head[:] = 0.0311, 0.1358, 0.0117
    bone.tail[:] = 0.0324, 0.1297, 0.0092
    bone.roll = -1.0029
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['r_palm.002.L']]
    bones['r_middle.001.L'] = bone.name
    bone = arm.edit_bones.new('r_ring.001.L')
    bone.head[:] = 0.0376, 0.1372, 0.0117
    bone.tail[:] = 0.0389, 0.1311, 0.0092
    bone.roll = -1.0029
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['r_palm.003.L']]
    bones['r_ring.001.L'] = bone.name
    bone = arm.edit_bones.new('r_pinky.001.L')
    bone.head[:] = 0.0466, 0.1444, 0.0083
    bone.tail[:] = 0.0476, 0.1412, 0.0074
    bone.roll = -1.7551
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['r_palm.004.L']]
    bones['r_pinky.001.L'] = bone.name
    bone = arm.edit_bones.new('r_index.002.L')
    bone.head[:] = 0.0217, 0.1325, 0.0070
    bone.tail[:] = 0.0221, 0.1271, 0.0038
    bone.roll = -0.2465
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['r_index.001.L']]
    bones['r_index.002.L'] = bone.name
    bone = arm.edit_bones.new('r_middle.002.L')
    bone.head[:] = 0.0324, 0.1297, 0.0092
    bone.tail[:] = 0.0343, 0.1210, 0.0039
    bone.roll = -0.7479
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['r_middle.001.L']]
    bones['r_middle.002.L'] = bone.name
    bone = arm.edit_bones.new('r_ring.002.L')
    bone.head[:] = 0.0389, 0.1311, 0.0092
    bone.tail[:] = 0.0407, 0.1229, 0.0042
    bone.roll = -0.7479
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['r_ring.001.L']]
    bones['r_ring.002.L'] = bone.name
    bone = arm.edit_bones.new('r_pinky.002.L')
    bone.head[:] = 0.0476, 0.1412, 0.0074
    bone.tail[:] = 0.0494, 0.1351, 0.0032
    bone.roll = -0.8965
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['r_pinky.001.L']]
    bones['r_pinky.002.L'] = bone.name

    bpy.ops.object.mode_set(mode='OBJECT')
    pbone = obj.pose.bones[bones['thigh.L']]
    pbone.rigify_type = 'limbs.super_limb'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    try:
        pbone.rigify_parameters.limb_type = "paw"
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.fk_layers = [False, False, False, False, False, False, False, False, False, False, False, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False]
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.tweak_layers = [False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False]
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.segments = 2
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['shin.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['foot.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['r_toe.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['r_palm.001.L']]
    pbone.rigify_type = 'limbs.super_palm'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['r_palm.002.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['r_palm.003.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['r_palm.004.L']]
    pbone.rigify_type = 'limbs.super_palm'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['r_index.001.L']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    try:
        pbone.rigify_parameters.tweak_layers = [False, False, False, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False]
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['r_middle.001.L']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    try:
        pbone.rigify_parameters.tweak_layers = [False, False, False, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False]
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['r_ring.001.L']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    try:
        pbone.rigify_parameters.tweak_layers = [False, False, False, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False]
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['r_pinky.001.L']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    try:
        pbone.rigify_parameters.tweak_layers = [False, False, False, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False]
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['r_index.002.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['r_middle.002.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['r_ring.002.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['r_pinky.002.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'

    bpy.ops.object.mode_set(mode='EDIT')
    for bone in arm.edit_bones:
        bone.select = False
        bone.select_head = False
        bone.select_tail = False
    for b in bones:
        bone = arm.edit_bones[bones[b]]
        bone.select = True
        bone.select_head = True
        bone.select_tail = True
        arm.edit_bones.active = bone

    for eb in arm.edit_bones:
        if ('thigh' in eb.name) or ('shin' in eb.name) or ('foot' in eb.name) or ('toe' in eb.name):
            eb.layers = (False, False, False, False, False, False, False, False, False, False, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False)
        else:
            eb.layers = (False, False, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False)
    arm.layers = (False, False, False, False, False, True, False, False, False, False, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False)


if __name__ == "__main__":
    create_sample(bpy.context.active_object)



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####


bl_info = {
    "name": "Edit Operator Source",
    "author": "scorpion81",
    "version": (1, 2, 2),
    "blender": (2, 78, 0),
    "location": "Text Editor > Edit > Edit Operator",
    "description": "Opens source file of chosen operator, if it is an add-on one",
    "warning": "",
    "wiki_url": "https://wiki.blender.org/index.php/Extensions:2.6/"
                "Py/Scripts/Development/Edit_Operator_Source",
    "category": "Development"}

import bpy
import sys
import inspect
from bpy.types import (
        Operator,
        Panel,
        )
from bpy.props import EnumProperty


def getclazz(opname):
    opid = opname.split(".")
    opmod = getattr(bpy.ops, opid[0])
    op = getattr(opmod, opid[1])
    id = op.get_rna().bl_rna.identifier
    clazz = getattr(bpy.types, id)
    return clazz


def getmodule(opname):
    addon = True
    clazz = getclazz(opname)
    modn = clazz.__module__

    try:
        line = inspect.getsourcelines(clazz)[1]
    except IOError:
        line = -1
    except TypeError:
        line = -1

    if modn == 'bpy.types':
        mod = 'C operator'
        addon = False
    elif modn != '__main__':
        mod = sys.modules[modn].__file__
    else:
        addon = False
        mod = modn

    return mod, line, addon


def get_ops():
    allops = []
    opsdir = dir(bpy.ops)
    for opmodname in opsdir:
        opmod = getattr(bpy.ops, opmodname)
        opmoddir = dir(opmod)
        for o in opmoddir:
            name = opmodname + "." + o
            clazz = getclazz(name)
            if (clazz.__module__ != 'bpy.types'):
                allops.append(name)
        del opmoddir

    # add own operator name too, since its not loaded yet when this is called
    allops.append("text.edit_operator")
    l = sorted(allops)
    del allops
    del opsdir

    return [(y, y, "", x) for x, y in enumerate(l)]


class EditOperator(Operator):
    bl_idname = "text.edit_operator"
    bl_label = "Edit Operator"
    bl_description = "Opens the source file of operators chosen from Menu"
    bl_property = "op"

    items = get_ops()

    op = EnumProperty(
            name="Op",
            description="",
            items=items
            )

    def invoke(self, context, event):
        context.window_manager.invoke_search_popup(self)
        return {'PASS_THROUGH'}

    def execute(self, context):
        found = False
        path, line, addon = getmodule(self.op)
        if addon:
            for t in bpy.data.texts:
                if t.filepath == path:
                    ctx = context.copy()
                    ctx['edit_text'] = t
                    bpy.ops.text.jump(ctx, line=line)
                    found = True
                    break

            if (found is False):
                self.report({'INFO'},
                            "Opened file: " + path)
                bpy.ops.text.open(filepath=path)
                bpy.ops.text.jump(line=line)

            return {'FINISHED'}
        else:
            self.report({'WARNING'},
                        "Found no source file for " + self.op)

            return {'CANCELLED'}


class EditOperatorPanel(Panel):
    bl_space_type = 'TEXT_EDITOR'
    bl_region_type = 'UI'
    bl_label = "Edit Operator"

    def draw(self, context):
        layout = self.layout
        layout.operator("text.edit_operator")


def register():
    bpy.utils.register_class(EditOperator)
    bpy.utils.register_class(EditOperatorPanel)


def unregister():
    bpy.utils.unregister_class(EditOperatorPanel)
    bpy.utils.unregister_class(EditOperator)


if __name__ == "__main__":
    register()



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

bl_info = {
    "name": "Useless Tools",
    "description": "Just a little collection of scripts and tools I use daily",
    "author": "Greg Zaal",
    "version": (1, 2, 2),
    "blender": (2, 75, 0),
    "location": "3D View > Tools",
    "warning": "",
    "wiki_url": "",
    "category": "Tools"}


import bpy
from bpy.types import Operator
from bpy.props import BoolProperty


def error_handlers(self, op_name, errors, reports="ERROR"):
    if self and reports:
        self.report({'INFO'},
                    reports + ": some operations could not be performed "
                    "(See Console for more info)")

    str_errors = "\n".join(errors)
    print("\n[Display Tools]\nOperator: {}\nErrors: {}\n".format(op_name, str_errors))


class UTSetSelectable(Operator):
    bl_idname = "ut.set_selectable"
    bl_label = "Set Selectable"
    bl_description = "Sets selectability for the selected objects"

    selectable = BoolProperty()

    def execute(self, context):
        errors = []
        for obj in bpy.context.selected_objects:
            try:
                if self.selectable is True:
                    obj.hide_select = False
                else:
                    obj.hide_select = True
            except Exception as k:
                name = getattr(obj, "name", "Nameless")
                errors.append("Error on {} - {}".format(name, k))
        if errors:
            error_handlers(self, "ut.set_selectable", errors, "Set Selectable")

        return {'FINISHED'}


class UTSetRenderable(Operator):
    bl_idname = "ut.set_renderable"
    bl_label = "Set Renderable"
    bl_description = "Sets renderability for the selected objects"

    renderable = BoolProperty()

    def execute(self, context):
        errors = []
        for obj in bpy.context.selected_objects:
            try:
                if self.renderable is True:
                    obj.hide_render = False
                else:
                    obj.hide_render = True
            except Exception as k:
                name = getattr(obj, "name", "Nameless")
                errors.append("Error on {} - {}".format(name, k))
        if errors:
            error_handlers(self, "ut.set_renderable", errors, "Set Renderable")

        return {'FINISHED'}


class UTAllSelectable(Operator):
    bl_idname = "ut.all_selectable"
    bl_label = "All Selectable"
    bl_description = "Allows all objects to be selected"

    def execute(self, context):
        errors = []
        for obj in bpy.data.objects:
            try:
                obj.hide_select = False
            except Exception as k:
                name = getattr(obj, "name", "Nameless")
                errors.append("Error on {} - {}".format(name, k))
        if errors:
            error_handlers(self, "ut.all_selectable", errors, "All Selectable")

        return {'FINISHED'}


class UTAllRenderable(Operator):
    bl_idname = "ut.all_renderable"
    bl_label = "All Renderable"
    bl_description = "Allow all objects to be rendered"

    def execute(self, context):
        errors = []
        for obj in bpy.data.objects:
            try:
                obj.hide_render = False
            except Exception as k:
                name = getattr(obj, "name", "Nameless")
                errors.append("Error on {} - {}".format(name, k))
        if errors:
            error_handlers(self, "ut.all_renderable", errors, "All Renderable")

        return {'FINISHED'}


class UTSelNGon(Operator):
    bl_idname = "ut.select_ngons"
    bl_label = "Select NGons"
    bl_description = "Select faces with more than 4 vertices"

    @classmethod
    def poll(cls, context):
        if not context.active_object or context.mode != 'EDIT_MESH':
            return False
        return True

    def execute(self, context):
        errors = []
        try:
            context.tool_settings.mesh_select_mode = (False, False, True)
            bpy.ops.mesh.select_face_by_sides(number=4, type='GREATER', extend=True)
        except Exception as k:
            errors.append("Error - {}".format(k))
        if errors:
            error_handlers(self, "ut.select_ngons", errors, "Select NGons")

        return {'FINISHED'}


class UTWireShowHideSelAll(Operator):
    bl_idname = "ut.wire_show_hide"
    bl_label = "Show / Hide Wire Selected or All"
    bl_description = "Change the status of the Wire display on Selected Objects"

    show = BoolProperty(
            default=False
            )
    selected = BoolProperty(
            default=False
            )

    @classmethod
    def poll(cls, context):
        return not context.scene.display_tools.WT_handler_enable

    def execute(self, context):
        errors = []
        objects = bpy.context.selected_objects if self.selected else bpy.data.objects
        for e in objects:
            try:
                e.show_wire = self.show
            except Exception as k:
                name = getattr(e, "name", "Nameless")
                errors.append("Error on {} - {}".format(name, k))
        if errors:
            error_handlers(self, "ut.wire_show_hide", errors,
                           "Show / Hide Wire Selected or All")

        return {'FINISHED'}


class UTSubsurfHideSelAll(Operator):
    bl_idname = "ut.subsurf_show_hide"
    bl_label = "Subsurf Show/Hide"
    bl_description = ("Sets the Subsurf modifier on objects:\n"
                      "Hide and Show operate on Selected Objects only\n"
                      "Hide All and Show All operate on All Objects in the data")

    show = BoolProperty(
            default=False
            )
    selected = BoolProperty(
            default=False
            )

    def execute(self, context):
        errors = []
        objects = bpy.context.selected_objects if self.selected else bpy.data.objects
        for e in objects:
            try:
                if e.type not in {"LAMP", "CAMERA", "EMPTY"}:
                    e.modifiers['Subsurf'].show_viewport = self.show
            except Exception as k:
                name = getattr(e, "name", "Nameless")
                errors.append(
                    "No subsurf on {} or it is not named Subsurf\nError: {}".format(name, k))
        if errors:
            error_handlers(self, "ut.subsurf_show_hide", errors, "Subsurf Show/Hide")

        return {'FINISHED'}


class UTOptimalDisplaySelAll(Operator):
    bl_idname = "ut.optimaldisplay"
    bl_label = "Optimal Display"
    bl_description = "Disables Optimal Display for all Subsurf modifiers on objects"

    on = BoolProperty(
            default=False
            )
    selected = BoolProperty(
            default=False
            )

    def execute(self, context):
        errors = []
        objects = bpy.context.selected_objects if self.selected else bpy.data.objects
        for e in objects:
            try:
                if e.type not in {"LAMP", "CAMERA", "EMPTY"}:
                    e.modifiers['Subsurf'].show_only_control_edges = self.on
            except Exception as k:
                name = getattr(e, "name", "Nameless")
                errors.append(
                    "No subsurf on {} or it is not named Subsurf\nError: {}".format(name, k))
        if errors:
            error_handlers(self, "ut.optimaldisplay", errors, "Optimal Display")

        return {'FINISHED'}


class UTAllEdges(Operator):
    bl_idname = "ut.all_edges"
    bl_label = "All Edges"
    bl_description = "Change the status of All Edges overlay on all objects"

    on = BoolProperty(
            default=False
            )

    def execute(self, context):
        errors = []
        for e in bpy.data.objects:
            try:
                e.show_all_edges = self.on
            except Exception as k:
                name = getattr(e, "name", "Nameless")
                errors.append(
                    "Enabling All Edges  on {} \nError: {}".format(name, k))
        if errors:
            error_handlers(self, "ut.all_edges", errors, "All Edges")

        return {'FINISHED'}


class UTDoubleSided(Operator):
    bl_idname = "ut.double_sided"
    bl_label = "Double Sided Normals"
    bl_description = "Disables Double Sided Normals for all objects"

    on = BoolProperty(
            default=False
            )

    def execute(self, context):
        errors = []
        for e in bpy.data.meshes:
            try:
                e.show_double_sided = self.on
            except Exception as k:
                name = getattr(e, "name", "Nameless")
                errors.append(
                    "Applying Double Sided Normals on {} \nError: {}".format(name, k))
        if errors:
            error_handlers(self, "ut.double_sided", errors, "Double Sided Normals")

        return {'FINISHED'}


# Register
def register():
    bpy.utils.register_module(__name__)


def unregister():
    bpy.utils.unregister_module(__name__)


if __name__ == "__main__":
    register()



# This file is a part of the HiRISE DTM Importer for Blender
#
# Copyright (C) 2017 Arizona Board of Regents on behalf of the Planetary Image
# Research Laboratory, Lunar and Planetary Laboratory at the University of
# Arizona.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""PVL Label Parsing"""

import collections
import re

from . import patterns
from .label import Label

Quantity = collections.namedtuple('Quantity', ['value', 'units'])


class PVLParseError(Exception):
    """Error parsing PVL file"""
    def __init__(self, message):
        super(PVLParseError, self).__init__(message)


class LabelParser:
    """A PVL Parser"""
    @staticmethod
    def load(path):
        """
        Load a dict-like representation of a PVL label header

        Parameters
        ----------
        path : str
            Path to a file containing a PVL header

        Returns
        ----------
        label : pvl.Label

        """
        raw = LabelParser._read(path)
        return Label(**LabelParser._parse(raw))

    @staticmethod
    def _read(path):
        """
        Get the PVL header from a file as a string

        Parameters
        ----------
        path : str
            Path to a file containing a PVL header

        Returns
        ----------
        raw : str

        Notes
        ---------
        * This function assumes that the file begins with a PVL header
          and it will read lines from the file until it encounters
          a PVL end statement.

        To-Do
        ---------
        * This could be more robust. What happens if there is no label
          in the file?

        """
        with open(path, 'rb') as f:
            raw = ''
            while True:
                try:
                    line = f.readline().decode()
                    raw += line
                    if re.match(patterns.END, line):
                        break
                except UnicodeDecodeError:
                    raise PVLParseError("Error parsing PVL label from "
                                        "file: {}".format(path))
        return raw

    @staticmethod
    def _remove_comments(raw):
        return re.sub(patterns.COMMENT, '', raw)

    @staticmethod
    def _parse(raw):
        raw = LabelParser._remove_comments(raw)
        label_iter = re.finditer(patterns.STATEMENT, raw)
        return LabelParser._parse_iter(label_iter)

    @staticmethod
    def _parse_iter(label_iter):
        """Recursively parse a PVL label iterator"""
        obj = {}
        while True:
            try:
                # Try to fetch the next match from the iter
                match = next(label_iter)
                val = match.group('val')
                key = match.group('key')
                # Handle nested object groups
                if key == 'OBJECT':
                    obj.update({
                        val: LabelParser._parse_iter(label_iter)
                    })
                elif key == 'END_OBJECT':
                    return obj
                # Add key/value pair to dict
                else:
                    # Should this value be a numeric type?
                    try:
                        val = LabelParser._convert_to_numeric(val)
                    except ValueError:
                        pass
                    # Should this value have units?
                    if match.group('units'):
                        val = Quantity(val, match.group('units'))
                    # Add it to the dict
                    obj.update({key: val})
            except StopIteration:
                break
        return obj

    @staticmethod
    def _convert_to_numeric(s):
        """Convert a string to its appropriate numeric type"""
        if re.match(patterns.INTEGER, s):
            return int(s)
        elif re.match(patterns.FLOATING, s):
            return float(s)
        else:
            raise ValueError



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8 compliant>

bl_info = {
    "name": "3D Print Toolbox",
    "author": "Campbell Barton",
    "blender": (2, 65, 0),
    "location": "3D View > Toolbox",
    "description": "Utilities for 3D printing",
    "warning": "",
    "wiki_url": "https://wiki.blender.org/index.php/Extensions:2.6/Py/"
                "Scripts/Modeling/PrintToolbox",
    "support": 'OFFICIAL',
    "category": "Mesh"}


if "bpy" in locals():
    import importlib
    importlib.reload(ui)
    importlib.reload(operators)
    importlib.reload(mesh_helpers)
else:
    import bpy
    from bpy.props import (
            StringProperty,
            BoolProperty,
            FloatProperty,
            EnumProperty,
            PointerProperty,
            )
    from bpy.types import (
            AddonPreferences,
            PropertyGroup,
            )
    from . import (
            ui,
            operators,
            )

import math


class Print3DSettings(PropertyGroup):
    export_format = EnumProperty(
            name="Format",
            description="Format type to export to",
            items=(('STL', "STL", ""),
                   ('PLY', "PLY", ""),
                   ('WRL', "VRML2", ""),
                   ('X3D', "X3D", ""),
                   ('OBJ', "OBJ", "")),
            default='STL',
            )
    use_export_texture = BoolProperty(
            name="Copy Textures",
            description="Copy textures on export to the output path",
            default=False,
            )
    use_apply_scale = BoolProperty(
            name="Apply Scale",
            description="Apply scene scale setting on export",
            default=False,
            )
    export_path = StringProperty(
            name="Export Directory",
            description="Path to directory where the files are created",
            default="//", maxlen=1024, subtype="DIR_PATH",
            )
    thickness_min = FloatProperty(
            name="Thickness",
            description="Minimum thickness",
            subtype='DISTANCE',
            default=0.001,  # 1mm
            min=0.0, max=10.0,
            )
    threshold_zero = FloatProperty(
            name="Threshold",
            description="Limit for checking zero area/length",
            default=0.0001,
            precision=5,
            min=0.0, max=0.2,
            )
    angle_distort = FloatProperty(
            name="Angle",
            description="Limit for checking distorted faces",
            subtype='ANGLE',
            default=math.radians(45.0),
            min=0.0, max=math.radians(180.0),
            )
    angle_sharp = FloatProperty(
            name="Angle",
            subtype='ANGLE',
            default=math.radians(160.0),
            min=0.0, max=math.radians(180.0),
            )
    angle_overhang = FloatProperty(
            name="Angle",
            subtype='ANGLE',
            default=math.radians(45.0),
            min=0.0, max=math.radians(90.0),
            )


# Add-ons Preferences Update Panel

# Define Panel classes for updating
panels = (
    ui.Print3DToolBarObject,
    ui.Print3DToolBarMesh,
    )


def update_panel(self, context):
    message = "3D Print Toolbox: Updating Panel locations has failed"
    try:
        for panel in panels:
            if "bl_rna" in panel.__dict__:
                bpy.utils.unregister_class(panel)

        for panel in panels:
            panel.bl_category = context.user_preferences.addons[__name__].preferences.category
            bpy.utils.register_class(panel)

    except Exception as e:
        print("\n[{}]\n{}\n\nError:\n{}".format(__name__, message, e))
        pass


class printpreferences(AddonPreferences):
    # this must match the addon name, use '__package__'
    # when defining this in a submodule of a python package.
    bl_idname = __name__

    category = StringProperty(
                name="Tab Category",
                description="Choose a name for the category of the panel",
                default="3D Printing",
                update=update_panel
                )

    def draw(self, context):
        layout = self.layout
        row = layout.row()
        col = row.column()

        col.label(text="Tab Category:")
        col.prop(self, "category", text="")


classes = (
    ui.Print3DToolBarObject,
    ui.Print3DToolBarMesh,

    operators.Print3DInfoVolume,
    operators.Print3DInfoArea,

    operators.Print3DCheckDegenerate,
    operators.Print3DCheckDistorted,
    operators.Print3DCheckSolid,
    operators.Print3DCheckIntersections,
    operators.Print3DCheckThick,
    operators.Print3DCheckSharp,
    operators.Print3DCheckOverhang,
    operators.Print3DCheckAll,

    operators.Print3DCleanIsolated,
    operators.Print3DCleanDistorted,
    # operators.Print3DCleanThin,
    operators.Print3DCleanNonManifold,

    operators.Print3DSelectReport,

    operators.Print3DScaleToVolume,
    operators.Print3DScaleToBounds,

    operators.Print3DExport,

    Print3DSettings,
    printpreferences,
    )


def register():
    for cls in classes:
        bpy.utils.register_class(cls)

    bpy.types.Scene.print_3d = PointerProperty(type=Print3DSettings)

    update_panel(None, bpy.context)


def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)

    del bpy.types.Scene.print_3d



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

import bpy
from bpy.types import (
        Panel,
        PropertyGroup,
        )
from bpy.props import (
        FloatProperty,
        EnumProperty,
        IntProperty,
        BoolProperty,
        FloatVectorProperty,
        )


# Class to define properties
class TracerProperties(PropertyGroup):
    """Options for tools"""
    curve_spline = EnumProperty(
            name="Spline",
            items=(("POLY", "Poly", "Use Poly spline type"),
                   ("NURBS", "Nurbs", "Use Nurbs spline type"),
                   ("BEZIER", "Bezier", "Use Bezier spline type")),
            description="Choose which type of spline to use when curve is created",
            default="BEZIER"
            )
    curve_handle = EnumProperty(
            name="Handle",
            items=(("ALIGNED", "Aligned", "Use Aligned Handle Type"),
                   ("AUTOMATIC", "Automatic", "Use Auto Handle Type"),
                   ("FREE_ALIGN", "Free Align", "Use Free Handle Type"),
                   ("VECTOR", "Vector", "Use Vector Handle Type")),
            description="Choose which type of handle to use when curve is created",
            default="VECTOR"
            )
    curve_resolution = IntProperty(
            name="Bevel Resolution",
            min=1, max=32,
            default=4,
            description="Adjust the Bevel resolution"
            )
    curve_depth = FloatProperty(
            name="Bevel Depth",
            min=0.0, max=100.0,
            default=0.1,
            description="Adjust the Bevel depth"
            )
    curve_u = IntProperty(
            name="Resolution U",
            min=0, max=64,
            default=12,
            description="Adjust the Surface resolution"
            )
    curve_join = BoolProperty(
            name="Join Curves",
            default=False,
            description="Join all the curves after they have been created"
            )
    curve_smooth = BoolProperty(
            name="Smooth",
            default=True,
            description="Render curve smooth"
            )
    # Option to Duplicate Mesh
    object_duplicate = BoolProperty(
            name="Apply to Copy",
            default=False,
            description="Apply curve to a copy of object"
            )
    # Distort Mesh options
    distort_modscale = IntProperty(
            name="Modulation Scale",
            min=0, max=50,
            default=2,
            description="Add a scale to modulate the curve at random points, set to 0 to disable"
            )
    distort_noise = FloatProperty(
            name="Mesh Noise",
            min=0.0, max=50.0,
            default=0.00,
            description="Adjust noise added to mesh before adding curve"
            )
    # Particle Options
    particle_step = IntProperty(
            name="Step Size",
            min=1, max=50,
            default=5,
            description="Sample one every this number of frames"
            )
    particle_auto = BoolProperty(
            name="Auto Frame Range",
            default=True,
            description="Calculate Frame Range from particles life"
            )
    particle_f_start = IntProperty(
            name='Start Frame',
            min=1, max=5000,
            default=1,
            description='Start frame'
            )
    particle_f_end = IntProperty(
            name="End Frame",
            min=1, max=5000,
            default=250,
            description="End frame"
            )
    # F-Curve Modifier Properties
    fcnoise_rot = BoolProperty(
            name="Rotation",
            default=False,
            description="Affect Rotation"
            )
    fcnoise_loc = BoolProperty(
            name="Location",
            default=True,
            description="Affect Location"
            )
    fcnoise_scale = BoolProperty(
            name="Scale",
            default=False,
            description="Affect Scale"
            )
    fcnoise_amp = IntProperty(
            name="Amp",
            min=1, max=500,
            default=5,
            description="Adjust the amplitude"
            )
    fcnoise_timescale = FloatProperty(
            name="Time Scale",
            min=1, max=500,
            default=50,
            description="Adjust the time scale"
            )
    fcnoise_key = BoolProperty(
            name="Add Keyframe",
            default=True,
            description="Keyframe is needed for tool, this adds a LocRotScale keyframe"
            )
    show_curve_settings = BoolProperty(
            name="Curve Settings",
            default=False,
            description="Change the curve settings for the created curve"
            )
    material_settings = BoolProperty(
            name="Material Settings",
            default=False,
            description="Change the material settings for the created curve"
            )
    particle_settings = BoolProperty(
            name="Particle Settings",
            default=False,
            description="Show the settings for the created curve"
            )
    animation_settings = BoolProperty(
            name="Animation Settings",
            default=False,
            description="Show the settings for the Animations"
            )
    distort_curve = BoolProperty(
            name="Add Distortion",
            default=False,
            description="Set options to distort the final curve"
            )
    connect_noise = BoolProperty(
            name="F-Curve Noise",
            default=False,
            description="Adds F-Curve Noise Modifier to selected objects"
            )
    settings_objectTrace = BoolProperty(
            name="Object Trace Settings",
            default=False,
            description="Trace selected mesh object with a curve"
            )
    settings_objectsConnect = BoolProperty(
            name="Objects Connect Settings",
            default=False,
            description="Connect objects with a curve controlled by hooks"
            )
    settings_objectTrace = BoolProperty(
            name="Object Trace Settings",
            default=False,
            description="Trace selected mesh object with a curve"
            )
    respect_order = BoolProperty(
            name="Order",
            default=False,
            description="Remember order objects were selected"
            )
    settings_particleTrace = BoolProperty(
            name="Particle Trace Settings",
            default=False,
            description="Trace particle path with a  curve"
            )
    settings_particleConnect = BoolProperty(
            name="Particle Connect Settings",
            default=False,
            description="Connect particles with a curves and animated over particle lifetime"
            )
    settings_growCurve = BoolProperty(
            name="Grow Curve Settings",
            default=False,
            description="Animate curve bevel over time by keyframing points radius"
            )
    settings_fcurve = BoolProperty(
            name="F-Curve Settings",
            default=False,
            description="F-Curve Settings"
            )
    settings_toggle = BoolProperty(
            name="Settings",
            default=False,
            description="Toggle Settings"
            )
    # Animation Options
    anim_auto = BoolProperty(
            name="Auto Frame Range",
            default=True,
            description="Automatically calculate Frame Range"
            )
    anim_f_start = IntProperty(
            name="Start",
            min=1, max=2500,
            default=1,
            description="Start frame / Hidden object"
            )
    anim_length = IntProperty(
            name="Duration",
            min=1,
            soft_max=1000, max=2500,
            default=100,
            description="Animation Length"
            )
    anim_f_fade = IntProperty(
            name="Fade After",
            min=0,
            soft_max=250, max=2500,
            default=10,
            description="Fade after this frames / Zero means no fade"
            )
    anim_delay = IntProperty(
            name="Grow",
            min=0, max=50,
            default=5,
            description="Frames it takes a point to grow"
            )
    anim_tails = BoolProperty(
            name='Tails on endpoints',
            default=True,
            description='Set radius to zero for open splines endpoints'
            )
    anim_keepr = BoolProperty(
            name="Keep Radius",
            default=True,
            description="Try to keep radius data from original curve"
            )
    animate = BoolProperty(
            name="Animate Result",
            default=False,
            description="Animate the final curve objects"
            )
    # Convert to Curve options
    convert_conti = BoolProperty(
            name="Continuous",
            default=True,
            description="Create a continuous curve using verts from mesh"
            )
    convert_everyedge = BoolProperty(
            name="Every Edge",
            default=False,
            description="Create a curve from all verts in a mesh"
            )
    convert_edgetype = EnumProperty(
            name="Edge Type for Curves",
            items=(("CONTI", "Continuous", "Create a continuous curve using verts from mesh"),
                   ("EDGEALL", "All Edges", "Create a curve from every edge in a mesh")),
            description="Choose which type of spline to use when curve is created",
            default="CONTI"
            )
    convert_joinbefore = BoolProperty(
            name="Join objects before convert",
            default=False,
            description="Join all selected mesh to one object before converting to mesh"
            )
    # Mesh Follow Options
    fol_edge_select = BoolProperty(
            name="Edge",
            default=False,
            description="Grow from edges"
            )
    fol_vert_select = BoolProperty(
            name="Vertex",
            default=False,
            description="Grow from verts"
            )
    fol_face_select = BoolProperty(
            name="Face",
            default=True,
            description="Grow from faces"
            )
    fol_mesh_type = EnumProperty(
            name="Mesh type",
            default="VERTS",
            description="Mesh feature to draw cruves from",
            items=(("VERTS", "Verts", "Draw from Verts"),
                   ("EDGES", "Edges", "Draw from Edges"),
                   ("FACES", "Faces", "Draw from Faces"),
                   ("OBJECT", "Object", "Draw from Object origin"))
            )
    fol_start_frame = IntProperty(
            name="Start Frame",
            min=1, max=2500,
            default=1,
            description="Start frame for range to trace"
            )
    fol_end_frame = IntProperty(
            name="End Frame",
            min=1, max=2500,
            default=250,
            description="End frame for range to trace"
            )
    fol_perc_verts = FloatProperty(
            name="Reduce selection by",
            min=0.001, max=1.000,
            default=0.5,
            description="percentage of total verts to trace"
            )
    fol_sel_option = EnumProperty(
            name="Selection type",
            description="Choose which objects to follow",
            default="RANDOM",
            items=(("RANDOM", "Random", "Follow Random items"),
                   ("CUSTOM", "Custom Select", "Follow selected items"),
                   ("ALL", "All", "Follow all items"))
            )
    trace_mat_color = FloatVectorProperty(
            name="Material Color",
            description="Choose material color",
            min=0, max=1,
            default=(0.0, 0.3, 0.6),
            subtype="COLOR"
            )
    trace_mat_random = BoolProperty(
            name="Random Color",
            default=False,
            description='Make the material colors random'
            )
    # Material custom Properties properties
    mat_simple_adv_toggle = EnumProperty(
            name="Material Options",
            items=(("SIMPLE", "Simple", "Show Simple Material Options"),
                   ("ADVANCED", "Advanced", "Show Advanced Material Options")),
            description="Choose which Material Options to show",
            default="SIMPLE"
            )
    mat_run_color_blender = BoolProperty(
            name="Run Color Blender",
            default=False,
            description="Generate colors from a color scheme"
            )
    mmColors = EnumProperty(
            items=(("RANDOM", "Random", "Use random colors"),
                    ("CUSTOM", "Custom", "Use custom colors"),
                    ("BW", "Black/White", "Use Black and White"),
                    ("BRIGHT", "Bright Colors", "Use Bright colors"),
                    ("EARTH", "Earth", "Use Earth colors"),
                    ("GREENBLUE", "Green to Blue", "Use Green to Blue colors")),
            description="Choose which type of colors the materials uses",
            default="BRIGHT",
            name="Define a color palette"
            )
    # Custom property for how many keyframes to skip
    mmSkip = IntProperty(
            name="frames",
            min=1, max=500,
            default=20,
            description="Number of frames between each keyframes"
            )
    # Custom property to enable/disable random order for the
    mmBoolRandom = BoolProperty(
            name="Random Order",
            default=False,
            description="Randomize the order of the colors"
            )
    # Custom Color properties
    mmColor1 = FloatVectorProperty(
            min=0, max=1,
            default=(0.8, 0.8, 0.8),
            description="Custom Color 1", subtype="COLOR"
            )
    mmColor2 = FloatVectorProperty(
            min=0, max=1,
            default=(0.8, 0.8, 0.3),
            description="Custom Color 2",
            subtype="COLOR"
            )
    mmColor3 = FloatVectorProperty(
            min=0, max=1,
            default=(0.8, 0.5, 0.6),
            description="Custom Color 3",
            subtype="COLOR"
            )
    mmColor4 = FloatVectorProperty(
            min=0, max=1,
            default=(0.2, 0.8, 0.289),
            description="Custom Color 4",
            subtype="COLOR"
            )
    mmColor5 = FloatVectorProperty(
            min=0, max=1,
            default=(1.0, 0.348, 0.8),
            description="Custom Color 5",
            subtype="COLOR"
            )
    mmColor6 = FloatVectorProperty(
            min=0, max=1,
            default=(0.4, 0.67, 0.8),
            description="Custom Color 6",
            subtype="COLOR"
            )
    mmColor7 = FloatVectorProperty(
            min=0, max=1,
            default=(0.66, 0.88, 0.8),
            description="Custom Color 7",
            subtype="COLOR"
            )
    mmColor8 = FloatVectorProperty(
            min=0, max=1,
            default=(0.8, 0.38, 0.22),
            description="Custom Color 8",
            subtype="COLOR"
            )
    # BW Color properties
    bwColor1 = FloatVectorProperty(
            min=0, max=1,
            default=(0.0, 0.0, 0.0),
            description="Black/White Color 1",
            subtype="COLOR"
            )
    bwColor2 = FloatVectorProperty(
            min=0, max=1,
            default=(1.0, 1.0, 1.0),
            description="Black/White Color 2",
            subtype="COLOR"
            )
    # Bright Color properties
    brightColor1 = FloatVectorProperty(
            min=0, max=1,
            default=(1.0, 0.0, 0.75),
            description="Bright Color 1",
            subtype="COLOR"
            )
    brightColor2 = FloatVectorProperty(
            min=0, max=1,
            default=(0.0, 1.0, 1.0),
            description="Bright Color 2",
            subtype="COLOR"
            )
    brightColor3 = FloatVectorProperty(
            min=0, max=1,
            default=(0.0, 1.0, 0.0),
            description="Bright Color 3",
            subtype="COLOR"
            )
    brightColor4 = FloatVectorProperty(
            min=0, max=1,
            default=(1.0, 1.0, 0.0),
            description="Bright Color 4", subtype="COLOR"
            )
    # Earth Color Properties
    earthColor1 = FloatVectorProperty(
            min=0, max=1,
            default=(0.068, 0.019, 0.014),
            description="Earth Color 1",
            subtype="COLOR"
            )
    earthColor2 = FloatVectorProperty(
            min=0, max=1,
            default=(0.089, 0.060, 0.047),
            description="Earth Color 2",
            subtype="COLOR"
            )
    earthColor3 = FloatVectorProperty(
            min=0, max=1,
            default=(0.188, 0.168, 0.066),
            description="Earth Color 3",
            subtype="COLOR"
            )
    earthColor4 = FloatVectorProperty(
            min=0, max=1,
            default=(0.445, 0.296, 0.065),
            description="Earth Color 4",
            subtype="COLOR"
            )
    earthColor5 = FloatVectorProperty(
            min=0, max=1,
            default=(0.745, 0.332, 0.065),
            description="Earth Color 5",
            subtype="COLOR"
            )
    # Green to Blue Color properties
    greenblueColor1 = FloatVectorProperty(
            min=0, max=1,
            default=(0.296, 0.445, 0.074),
            description="Green/Blue Color 1",
            subtype="COLOR"
            )
    greenblueColor2 = FloatVectorProperty(
            min=0, max=1,
            default=(0.651, 1.0, 0.223),
            description="Green/Blue Color 2",
            subtype="COLOR"
            )
    greenblueColor3 = FloatVectorProperty(
            min=0, max=1,
            default=(0.037, 0.047, 0.084),
            description="Green/Blue Color 3",
            subtype="COLOR"
            )

    # Toolbar show/hide booleans for tool options
    btrace_menu_items = [
            ('tool_help', "Help",
             "Pick one of the options below", "INFO", 0),
            ('tool_objectTrace', "Object Trace",
             "Trace selected mesh object with a curve", "FORCE_MAGNETIC", 1),
            ('tool_objectsConnect', "Objects Connect",
             "Connect objects with a curve controlled by hooks", "OUTLINER_OB_EMPTY", 2),
            ('tool_meshFollow', "Mesh Follow",
             "Follow selection items on animated mesh object", "DRIVER", 3),
            ('tool_handwrite', "Handwriting",
             "Create and Animate curve using the grease pencil", "BRUSH_DATA", 4),
            ('tool_particleTrace', "Particle Trace",
             "Trace particle path with a  curve", "PARTICLES", 5),
            ('tool_particleConnect', "Particle Connect",
             "Connect particles with a curves and animated over particle lifetime", "MOD_PARTICLES", 6),
            ('tool_growCurve', "Grow Curve",
             "Animate curve bevel over time by keyframing points radius", "META_BALL", 7),
            ('tool_fcurve', "F-Curve Noise",
             "Add F-Curve noise to selected objects", "RNDCURVE", 8),
            ('tool_colorblender', "Color Blender",
             "Pick the color of the created curves", "COLOR", 9),
            ]
    btrace_toolmenu = EnumProperty(
            name="Tools",
            items=btrace_menu_items,
            description="",
            default='tool_help'
            )


# Draw Brush panel in Toolbar
class addTracerObjectPanel(Panel):
    bl_label = "Btrace"
    bl_space_type = "VIEW_3D"
    bl_region_type = "TOOLS"
    bl_context = "objectmode"
    bl_category = "Create"
    bl_options = {'DEFAULT_CLOSED'}

    def draw(self, context):
        layout = self.layout
        Btrace = context.window_manager.curve_tracer
        addon_prefs = context.user_preferences.addons["btrace"].preferences
        switch_expand = addon_prefs.expand_enum
        obj = context.object

        # Color Blender Panel options
        def color_blender():
            # Buttons for Color Blender
            row = box.row()
            row.label("Color palette")
            row.prop(Btrace, "mmColors", text="")

            # Show Custom Colors if selected
            if Btrace.mmColors == 'CUSTOM':
                row = box.row(align=True)
                for i in range(1, 9):
                    row.prop(Btrace, "mmColor" + str(i), text="")
            # Show Earth Colors
            elif Btrace.mmColors == 'BW':
                row = box.row(align=True)
                row.prop(Btrace, "bwColor1", text="")
                row.prop(Btrace, "bwColor2", text="")
            # Show Earth Colors
            elif Btrace.mmColors == 'BRIGHT':
                row = box.row(align=True)
                for i in range(1, 5):
                    row.prop(Btrace, "brightColor" + str(i), text="")
            # Show Earth Colors
            elif Btrace.mmColors == 'EARTH':
                row = box.row(align=True)
                for i in range(1, 6):
                    row.prop(Btrace, "earthColor" + str(i), text="")
            # Show Earth Colors
            elif Btrace.mmColors == 'GREENBLUE':
                row = box.row(align=True)
                for i in range(1, 4):
                    row.prop(Btrace, "greenblueColor" + str(i), text="")
            elif Btrace.mmColors == 'RANDOM':
                row = box.row()

        # Curve noise settings
        def curve_noise():
            row = box.row()
            row.label(text="F-Curve Noise", icon='RNDCURVE')
            row = box.row(align=True)
            row.prop(Btrace, "fcnoise_rot", toggle=True)
            row.prop(Btrace, "fcnoise_loc", toggle=True)
            row.prop(Btrace, "fcnoise_scale", toggle=True)

            col = box.column(align=True)
            col.prop(Btrace, "fcnoise_amp")
            col.prop(Btrace, "fcnoise_timescale")
            box.prop(Btrace, "fcnoise_key")

        # Curve Panel options
        def curve_settings():
            # Button for curve options
            row = self.layout.row()
            row = box.row(align=True)

            row.prop(Btrace, "show_curve_settings",
                     icon='CURVE_BEZCURVE', text="Curve Settings")
            row.prop(Btrace, "material_settings",
                     icon='MATERIAL_DATA', text="Material Settings")

            if Btrace.material_settings:
                row = box.row()
                row.label(text="Material Settings", icon='COLOR')
                row = box.row()
                row.prop(Btrace, "trace_mat_random")
                if not Btrace.trace_mat_random:
                    row = box.row()
                    row.prop(Btrace, "trace_mat_color", text="")
                else:
                    row.prop(Btrace, "mat_run_color_blender")
                    if Btrace.mat_run_color_blender:
                        row = box.row()
                        row.operator("object.colorblenderclear",
                                     text="Reset Material Keyframes",
                                     icon="KEY_DEHLT")
                        row.prop(Btrace, "mmSkip", text="Keyframe every")
                    color_blender()
                row = box.row()

            if Btrace.show_curve_settings:
                # selected curve options
                if len(context.selected_objects) > 0 and obj.type == 'CURVE':
                    col = box.column(align=True)
                    col.label(text="Edit Curves for:", icon='IPO_BEZIER')
                    col.separator()
                    col.label(text="Selected Curve Bevel Options")
                    row = col.row(align=True)
                    row.prop(obj.data, "bevel_depth", text="Depth")
                    row.prop(obj.data, "bevel_resolution", text="Resolution")
                    row = col.row(align=True)
                    row.prop(obj.data, "resolution_u")
                else:  # For new curve
                    box.label(text="New Curve Settings", icon='CURVE_BEZCURVE')
                    box.prop(Btrace, "curve_spline")
                    box.prop(Btrace, "curve_handle")
                    box.label(text="Bevel Options")
                    col = box.column(align=True)
                    row = col.row(align=True)
                    row.prop(Btrace, "curve_depth", text="Depth")
                    row.prop(Btrace, "curve_resolution", text="Resolution")
                    row = col.row(align=True)
                    row.prop(Btrace, "curve_u")

        # Grow Animation Panel options
        def add_grow():
            # Button for grow animation option
            row = box.row()
            row.label(text="Animate Final Curve", icon="SPACE2")
            row = box.row()
            row.prop(Btrace, "animate", text="Add Grow Curve Animation", icon="META_BALL")
            box.separator()
            if Btrace.animate:
                box.label(text="Frame Animation Settings:", icon="META_BALL")
                col = box.column(align=True)
                col.prop(Btrace, "anim_auto")
                if not Btrace.anim_auto:
                    row = col.row(align=True)
                    row.prop(Btrace, "anim_f_start")
                    row.prop(Btrace, "anim_length")
                row = col.row(align=True)
                row.prop(Btrace, "anim_delay")
                row.prop(Btrace, "anim_f_fade")

                box.label(text="Additional Settings")
                row = box.row()
                row.prop(Btrace, "anim_tails")
                row.prop(Btrace, "anim_keepr")

        # Start Btrace Panel
        if switch_expand == 'list':
            layout.label(text="Available Tools:", icon="COLLAPSEMENU")
            col = layout.column(align=True)
            col.prop(Btrace, "btrace_toolmenu", text="")
        elif switch_expand == 'col':
            col = layout.column(align=True)
            col.prop(Btrace, "btrace_toolmenu", expand=True)
        elif switch_expand == 'row':
            row = layout.row(align=True)
            row.alignment = 'CENTER'
            row.prop(Btrace, "btrace_toolmenu", text="", expand=True)

        # Start Object Tools
        sel = context.selected_objects

        # Default option (can be expanded into help)
        if Btrace.btrace_toolmenu == 'tool_help':
            row = layout.row()
            row.label("Pick an option", icon="HELP")

        # Object Trace
        elif Btrace.btrace_toolmenu == 'tool_objectTrace':
            row = layout.row()
            row.label(text="  Trace Tool:", icon="FORCE_CURVE")
            box = self.layout.box()
            row = box.row()
            row.label(text="Object Trace", icon="FORCE_MAGNETIC")
            row.operator("object.btobjecttrace", text="Run!", icon="PLAY")
            row = box.row()
            row.prop(Btrace, "settings_toggle", icon="MODIFIER", text="Settings")
            myselected = "Selected %d" % len(context.selected_objects)
            row.label(text=myselected)
            if Btrace.settings_toggle:
                box.label(text="Edge Type for Curves:", icon="IPO_CONSTANT")
                row = box.row(align=True)
                row.prop(Btrace, "convert_edgetype", text="")
                box.prop(Btrace, "object_duplicate")
                if len(sel) > 1:
                    box.prop(Btrace, "convert_joinbefore")
                else:
                    Btrace.convert_joinbefore = False
                row = box.row()
                row.prop(Btrace, "distort_curve")
                if Btrace.distort_curve:
                    col = box.column(align=True)
                    col.prop(Btrace, "distort_modscale")
                    col.prop(Btrace, "distort_noise")
                row = box.row()
                curve_settings()  # Show Curve/material settings
                add_grow()        # Grow settings here

        # Objects Connect
        elif Btrace.btrace_toolmenu == 'tool_objectsConnect':
            row = layout.row()
            row.label(text="  Trace Tool:", icon="FORCE_CURVE")
            box = self.layout.box()
            row = box.row()
            row.label(text="Objects Connect", icon="OUTLINER_OB_EMPTY")
            row.operator("object.btobjectsconnect", text="Run!", icon="PLAY")
            row = box.row()
            row.prop(Btrace, "settings_toggle", icon='MODIFIER', text='Settings')
            row.label(text="")
            if Btrace.settings_toggle:
                row = box.row()
                row.prop(Btrace, "respect_order", text="Selection Options")
                if Btrace.respect_order:
                    box.operator("object.select_order",
                                 text="Click to start order selection",
                                 icon='UV_SYNC_SELECT')
                row = box.row()
                row.prop(Btrace, "connect_noise", text="Add F-Curve Noise")
                if Btrace.connect_noise:
                    curve_noise()     # Show Curve Noise settings

                curve_settings()      # Show Curve/material settings
                add_grow()            # Grow settings here

        # Mesh Follow
        elif Btrace.btrace_toolmenu == 'tool_meshFollow':
            row = layout.row()
            row.label(text="  Trace Tool:", icon="FORCE_CURVE")
            box = self.layout.box()
            row = box.row()
            row.label(text="Mesh Follow", icon="DRIVER")
            row.operator("object.btmeshfollow", text="Run!", icon="PLAY")
            row = box.row()
            if Btrace.fol_mesh_type == 'OBJECT':
                a, b = "Trace Object", "SNAP_VOLUME"
            if Btrace.fol_mesh_type == 'VERTS':
                a, b = "Trace Verts", "SNAP_VERTEX"
            if Btrace.fol_mesh_type == 'EDGES':
                a, b = "Trace Edges", "SNAP_EDGE"
            if Btrace.fol_mesh_type == 'FACES':
                a, b = "Trace Faces", "SNAP_FACE"
            row.prop(Btrace, "settings_toggle", icon='MODIFIER', text='Settings')
            row.label(text=a, icon=b)
            if Btrace.settings_toggle:
                col = box.column(align=True)
                row = col.row(align=True)
                row.prop(Btrace, "fol_mesh_type", expand=True)
                row = col.row(align=True)
                if Btrace.fol_mesh_type != 'OBJECT':
                    row.prop(Btrace, "fol_sel_option", expand=True)
                    row = box.row()
                    if Btrace.fol_sel_option == 'RANDOM':
                        row.label("Random Select of Total")
                        row.prop(Btrace, "fol_perc_verts", text="%")
                    if Btrace.fol_sel_option == 'CUSTOM':
                        row.label("Choose selection in Edit Mode")
                    if Btrace.fol_sel_option == 'ALL':
                        row.label("Select All items")
                col = box.column(align=True)
                col.label("Time Options", icon="TIME")
                col.prop(Btrace, "particle_step")
                row = col.row(align=True)
                row.prop(Btrace, "fol_start_frame")
                row.prop(Btrace, "fol_end_frame")
                curve_settings()  # Show Curve/material settings
                add_grow()        # Grow settings here

        # Handwriting Tools
        elif Btrace.btrace_toolmenu == 'tool_handwrite':
            row = layout.row()
            row.label(text="  Trace Tool:", icon="FORCE_CURVE")
            box = self.layout.box()
            row = box.row()
            row.label(text='Handwriting', icon='BRUSH_DATA')
            row.operator("curve.btwriting", text="Run!", icon='PLAY')
            row = box.row()
            row = box.row()
            row.label(text='Grease Pencil Writing Tools')
            col = box.column(align=True)
            row = col.row(align=True)
            row.operator("gpencil.draw", text="Draw", icon='BRUSH_DATA').mode = 'DRAW'
            row.operator("gpencil.draw", text="Poly", icon='VPAINT_HLT').mode = 'DRAW_POLY'
            row = col.row(align=True)
            row.operator("gpencil.draw", text="Line", icon='ZOOMOUT').mode = 'DRAW_STRAIGHT'
            row.operator("gpencil.draw", text="Erase", icon='TPAINT_HLT').mode = 'ERASER'
            row = box.row()
            row.operator("gpencil.data_unlink", text="Delete Grease Pencil Layer", icon="CANCEL")
            row = box.row()
            curve_settings()  # Show Curve/material settings
            add_grow()        # Grow settings here

        # Particle Trace
        elif Btrace.btrace_toolmenu == 'tool_particleTrace':
            row = layout.row()
            row.label(text="  Trace Tool:", icon="FORCE_CURVE")
            box = self.layout.box()
            row = box.row()
            row.label(text="Particle Trace", icon="PARTICLES")
            row.operator("particles.particletrace", text="Run!", icon="PLAY")
            row = box.row()
            row.prop(Btrace, "settings_toggle", icon='MODIFIER', text='Settings')
            row.label(text="")
            if Btrace.settings_toggle:
                box.prop(Btrace, "particle_step")
                row = box.row()
                row.prop(Btrace, "curve_join")
                curve_settings()  # Show Curve/material settings
                add_grow()        # Grow settings here

        # Connect Particles
        elif Btrace.btrace_toolmenu == 'tool_particleConnect':
            row = layout.row()
            row.label(text="  Trace Tool:", icon="FORCE_CURVE")
            box = self.layout.box()
            row = box.row()
            row.label(text='Particle Connect', icon='MOD_PARTICLES')
            row.operator("particles.connect", icon="PLAY", text='Run!')
            row = box.row()
            row.prop(Btrace, "settings_toggle", icon='MODIFIER', text='Settings')
            row.label(text="")
            if Btrace.settings_toggle:
                box.prop(Btrace, "particle_step")
                row = box.row()
                row.prop(Btrace, 'particle_auto')
                if not Btrace.particle_auto:
                    row = box.row(align=True)
                    row.prop(Btrace, 'particle_f_start')
                    row.prop(Btrace, 'particle_f_end')
                curve_settings()  # Show Curve/material settings
                add_grow()        # Grow settings here

        # Grow Animation
        elif Btrace.btrace_toolmenu == 'tool_growCurve':
            row = layout.row()
            row.label(text="  Curve Tool:", icon="OUTLINER_OB_CURVE")
            box = self.layout.box()
            row = box.row()
            row.label(text="Grow Curve", icon="META_BALL")
            row.operator("curve.btgrow", text="Run!", icon="PLAY")
            row = box.row()
            row.prop(Btrace, "settings_toggle", icon="MODIFIER", text="Settings")
            row.operator("object.btreset", icon="KEY_DEHLT")
            if Btrace.settings_toggle:
                box.label(text="Frame Animation Settings:")
                col = box.column(align=True)
                col.prop(Btrace, "anim_auto")
                if not Btrace.anim_auto:
                    row = col.row(align=True)
                    row.prop(Btrace, "anim_f_start")
                    row.prop(Btrace, "anim_length")
                row = col.row(align=True)
                row.prop(Btrace, "anim_delay")
                row.prop(Btrace, "anim_f_fade")

                box.label(text="Additional Settings")
                row = box.row()
                row.prop(Btrace, "anim_tails")
                row.prop(Btrace, "anim_keepr")

        # F-Curve Noise Curve
        elif Btrace.btrace_toolmenu == 'tool_fcurve':
            row = layout.row()
            row.label(text="  Curve Tool:", icon="OUTLINER_OB_CURVE")
            box = self.layout.box()
            row = box.row()
            row.label(text="F-Curve Noise", icon='RNDCURVE')
            row.operator("object.btfcnoise", icon='PLAY', text="Run!")
            row = box.row()
            row.prop(Btrace, "settings_toggle", icon='MODIFIER', text='Settings')
            row.operator("object.btreset", icon='KEY_DEHLT')
            if Btrace.settings_toggle:
                curve_noise()

        # Color Blender
        elif Btrace.btrace_toolmenu == 'tool_colorblender':
            row = layout.row()
            row.label(text="  Curve/Object Tool:", icon="OUTLINER_OB_CURVE")
            box = self.layout.box()
            row = box.row()
            row.label(text="Color Blender", icon="COLOR")
            row.operator("object.colorblender", icon='PLAY', text="Run!")
            row = box.row()
            row.operator("object.colorblenderclear", text="Reset Keyframes", icon="KEY_DEHLT")
            row.prop(Btrace, "mmSkip", text="Keyframe every")
            color_blender()



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8 compliant>

"""Replace default list-style menu for transform orientations with a pie."""

bl_info = {
    "name": "Hotkey: 'Alt + Spacebar'",
    "author": "Italic_",
    "version": (1, 1, 0),
    "blender": (2, 77, 0),
    "description": "Set Transform Orientations",
    "location": "3D View",
    "category": "Orientation Pie"}

import bpy
from bpy.types import (
        Menu,
        Operator,
        )
from bpy.props import (
        StringProperty,
        )


class OrientPoll(Operator):
    bl_idname = "pie.orientation"
    bl_label = "Orientation Poll"
    bl_options = {'INTERNAL'}

    space = StringProperty()

    @classmethod
    def poll(cls, context):
        return context.space_data.type == "VIEW_3D"

    def execute(self, context):
        context.space_data.transform_orientation = self.space
        return {'FINISHED'}


class OrientPie(Menu):
    bl_label = "Transform Orientation"
    bl_idname = "pie.orient"

    def draw(self, context):
        layout = self.layout
        pie = layout.menu_pie()
        view = context.space_data

        pie.operator("pie.orientation", text="Global").space = 'GLOBAL'
        pie.operator("pie.orientation", text="Local").space = 'LOCAL'
        pie.operator("pie.orientation", text="Gimbal").space = 'GIMBAL'

        # XXX: Display only custom orientations
        pie = pie.box()
        pie.prop(view, "transform_orientation", text="")
        pie = layout.menu_pie()

        pie.operator("pie.orientation", text="Normal").space = 'NORMAL'
        pie.operator("pie.orientation", text="View").space = 'VIEW'


addon_keymaps = []

classes = (
    OrientPie,
    OrientPoll
    )


def register():
    for cls in classes:
        bpy.utils.register_class(cls)

    wm = bpy.context.window_manager
    if wm.keyconfigs.addon:
        # Manipulators
        km = wm.keyconfigs.addon.keymaps.new(name='3D View Generic', space_type='VIEW_3D')
        kmi = km.keymap_items.new('wm.call_menu_pie', 'SPACE', 'PRESS', alt=True)
        kmi.properties.name = "pie.orient"
        addon_keymaps.append((km, kmi))


def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)

    wm = bpy.context.window_manager
    kc = wm.keyconfigs.addon
    if kc:
        for km, kmi in addon_keymaps:
            km.keymap_items.remove(kmi)
    addon_keymaps.clear()


if __name__ == "__main__":
    register()



# This file is a part of the HiRISE DTM Importer for Blender
#
# Copyright (C) 2017 Arizona Board of Regents on behalf of the Planetary Image
# Research Laboratory, Lunar and Planetary Laboratory at the University of
# Arizona.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""A HiRISE DTM Importer for Blender"""

bl_info = {
    "name": "HiRISE DTM Importer",
    "author": "Nicholas Wolf (nicwolf@pirl.lpl.arizona.edu)",
    "version": (0, 2, 2),
    "blender": (2, 78, 0),
    "location": "File > Import > HiRISE DTM (.img)",
    "description": "Import a HiRISE DTM as a mesh",
    "warning": "May consume a lot of memory",
    "wiki_url": "https://wiki.blender.org/index.php/Extensions:2.6/Py/"
                "Scripts/Import-Export/HiRISE_DTM_from_PDS_IMG",
    "category": "Import-Export",
}

if "bpy" in locals():
    import importlib
    importlib.reload(importer)
    importlib.reload(terrainpanel)

else:
    from .ui import importer
    from .ui import terrainpanel

import bpy


def menu_import(self, context):
    i = importer.ImportHiRISETerrain
    self.layout.operator(i.bl_idname, text=i.bl_label)


def register():
    bpy.utils.register_module(__name__)
    bpy.types.INFO_MT_file_import.append(menu_import)


def unregister():
    bpy.utils.unregister_module(__name__)
    bpy.types.INFO_MT_file_import.remove(menu_import)


if __name__ == '__main__':
    register()



# <pep8-80 compliant>

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Nutti <nutti.metro@gmail.com>"
__status__ = "production"
__version__ = "4.4"
__date__ = "2 Aug 2017"

from enum import IntEnum
import math

import bpy
import bgl
import mathutils
import bmesh

from . import muv_common


MAX_VALUE = 100000.0


class MUV_UVBBCmd():
    """
    Custom class: Base class of command
    """

    def __init__(self):
        self.op = 'NONE'        # operation

    def to_matrix(self):
        # mat = I
        mat = mathutils.Matrix()
        mat.identity()
        return mat


class MUV_UVBBTranslationCmd(MUV_UVBBCmd):
    """
    Custom class: Translation operation
    """

    def __init__(self, ix, iy):
        super().__init__()
        self.op = 'TRANSLATION'
        self.__x = ix       # current x
        self.__y = iy       # current y
        self.__ix = ix      # initial x
        self.__iy = iy      # initial y

    def to_matrix(self):
        # mat = Mt
        dx = self.__x - self.__ix
        dy = self.__y - self.__iy
        return mathutils.Matrix.Translation((dx, dy, 0))

    def set(self, x, y):
        self.__x = x
        self.__y = y


class MUV_UVBBRotationCmd(MUV_UVBBCmd):
    """
    Custom class: Rotation operation
    """

    def __init__(self, ix, iy, cx, cy):
        super().__init__()
        self.op = 'ROTATION'
        self.__x = ix       # current x
        self.__y = iy       # current y
        self.__cx = cx      # center of rotation x
        self.__cy = cy      # center of rotation y
        dx = self.__x - self.__cx
        dy = self.__y - self.__cy
        self.__iangle = math.atan2(dy, dx)      # initial rotation angle

    def to_matrix(self):
        # mat = Mt * Mr * Mt^-1
        dx = self.__x - self.__cx
        dy = self.__y - self.__cy
        angle = math.atan2(dy, dx) - self.__iangle
        mti = mathutils.Matrix.Translation((-self.__cx, -self.__cy, 0.0))
        mr = mathutils.Matrix.Rotation(angle, 4, 'Z')
        mt = mathutils.Matrix.Translation((self.__cx, self.__cy, 0.0))
        return mt * mr * mti

    def set(self, x, y):
        self.__x = x
        self.__y = y


class MUV_UVBBScalingCmd(MUV_UVBBCmd):
    """
    Custom class: Scaling operation
    """

    def __init__(self, ix, iy, ox, oy, dir_x, dir_y, mat):
        super().__init__()
        self.op = 'SCALING'
        self.__ix = ix          # initial x
        self.__iy = iy          # initial y
        self.__x = ix           # current x
        self.__y = iy           # current y
        self.__ox = ox          # origin of scaling x
        self.__oy = oy          # origin of scaling y
        self.__dir_x = dir_x    # direction of scaling x
        self.__dir_y = dir_y    # direction of scaling y
        self.__mat = mat
        # initial origin of scaling = M(to original transform) * (ox, oy)
        iov = mat * mathutils.Vector((ox, oy, 0.0))
        self.__iox = iov.x      # initial origin of scaling X
        self.__ioy = iov.y      # initial origin of scaling y

    def to_matrix(self):
        """
        mat = M(to original transform)^-1 * Mt(to origin) * Ms *
              Mt(to origin)^-1 * M(to original transform)
        """
        m = self.__mat
        mi = self.__mat.inverted()
        mtoi = mathutils.Matrix.Translation((-self.__iox, -self.__ioy, 0.0))
        mto = mathutils.Matrix.Translation((self.__iox, self.__ioy, 0.0))
        # every point must be transformed to origin
        t = m * mathutils.Vector((self.__ix, self.__iy, 0.0))
        tix, tiy = t.x, t.y
        t = m * mathutils.Vector((self.__ox, self.__oy, 0.0))
        tox, toy = t.x, t.y
        t = m * mathutils.Vector((self.__x, self.__y, 0.0))
        tx, ty = t.x, t.y
        ms = mathutils.Matrix()
        ms.identity()
        if self.__dir_x == 1:
            ms[0][0] = (tx - tox) * self.__dir_x / (tix - tox)
        if self.__dir_y == 1:
            ms[1][1] = (ty - toy) * self.__dir_y / (tiy - toy)
        return mi * mto * ms * mtoi * m

    def set(self, x, y):
        self.__x = x
        self.__y = y


class MUV_UVBBUniformScalingCmd(MUV_UVBBCmd):
    """
    Custom class: Uniform Scaling operation
    """

    def __init__(self, ix, iy, ox, oy, mat):
        super().__init__()
        self.op = 'SCALING'
        self.__ix = ix          # initial x
        self.__iy = iy          # initial y
        self.__x = ix           # current x
        self.__y = iy           # current y
        self.__ox = ox          # origin of scaling x
        self.__oy = oy          # origin of scaling y
        self.__mat = mat
        # initial origin of scaling = M(to original transform) * (ox, oy)
        iov = mat * mathutils.Vector((ox, oy, 0.0))
        self.__iox = iov.x      # initial origin of scaling x
        self.__ioy = iov.y      # initial origin of scaling y
        self.__dir_x = 1
        self.__dir_y = 1

    def to_matrix(self):
        """
        mat = M(to original transform)^-1 * Mt(to origin) * Ms *
              Mt(to origin)^-1 * M(to original transform)
        """
        m = self.__mat
        mi = self.__mat.inverted()
        mtoi = mathutils.Matrix.Translation((-self.__iox, -self.__ioy, 0.0))
        mto = mathutils.Matrix.Translation((self.__iox, self.__ioy, 0.0))
        # every point must be transformed to origin
        t = m * mathutils.Vector((self.__ix, self.__iy, 0.0))
        tix, tiy = t.x, t.y
        t = m * mathutils.Vector((self.__ox, self.__oy, 0.0))
        tox, toy = t.x, t.y
        t = m * mathutils.Vector((self.__x, self.__y, 0.0))
        tx, ty = t.x, t.y
        ms = mathutils.Matrix()
        ms.identity()
        tir = math.sqrt((tix - tox) * (tix - tox) + (tiy - toy) * (tiy - toy))
        tr = math.sqrt((tx - tox) * (tx - tox) + (ty - toy) * (ty - toy))

        sr = tr / tir

        if ((tx - tox) * (tix - tox)) > 0:
            self.__dir_x = 1
        else:
            self.__dir_x = -1
        if ((ty - toy) * (tiy - toy)) > 0:
            self.__dir_y = 1
        else:
            self.__dir_y = -1

        ms[0][0] = sr * self.__dir_x
        ms[1][1] = sr * self.__dir_y

        return mi * mto * ms * mtoi * m

    def set(self, x, y):
        self.__x = x
        self.__y = y


class MUV_UVBBCmdExecuter():
    """
    Custom class: manage command history and execute command
    """

    def __init__(self):
        self.__cmd_list = []        # history
        self.__cmd_list_redo = []   # redo list

    def execute(self, begin=0, end=-1):
        """
        create matrix from history
        """
        mat = mathutils.Matrix()
        mat.identity()
        for i, cmd in enumerate(self.__cmd_list):
            if begin <= i and (end == -1 or i <= end):
                mat = cmd.to_matrix() * mat
        return mat

    def undo_size(self):
        """
        get history size
        """
        return len(self.__cmd_list)

    def top(self):
        """
        get top of history
        """
        if len(self.__cmd_list) <= 0:
            return None
        return self.__cmd_list[-1]

    def append(self, cmd):
        """
        append command
        """
        self.__cmd_list.append(cmd)
        self.__cmd_list_redo = []

    def undo(self):
        """
        undo command
        """
        if len(self.__cmd_list) <= 0:
            return
        self.__cmd_list_redo.append(self.__cmd_list.pop())

    def redo(self):
        """
        redo command
        """
        if len(self.__cmd_list_redo) <= 0:
            return
        self.__cmd_list.append(self.__cmd_list_redo.pop())

    def pop(self):
        if len(self.__cmd_list) <= 0:
            return None
        return self.__cmd_list.pop()

    def push(self, cmd):
        self.__cmd_list.append(cmd)


class MUV_UVBBRenderer(bpy.types.Operator):
    """
    Operation class: Render UV bounding box
    """

    bl_idname = "uv.muv_uvbb_renderer"
    bl_label = "UV Bounding Box Renderer"
    bl_description = "Bounding Box Renderer about UV in Image Editor"

    __handle = None

    @staticmethod
    def handle_add(obj, context):
        if MUV_UVBBRenderer.__handle is None:
            sie = bpy.types.SpaceImageEditor
            MUV_UVBBRenderer.__handle = sie.draw_handler_add(
                MUV_UVBBRenderer.draw_bb,
                (obj, context), "WINDOW", "POST_PIXEL")

    @staticmethod
    def handle_remove():
        if MUV_UVBBRenderer.__handle is not None:
            sie = bpy.types.SpaceImageEditor
            sie.draw_handler_remove(
                MUV_UVBBRenderer.__handle, "WINDOW")
            MUV_UVBBRenderer.__handle = None

    @staticmethod
    def __draw_ctrl_point(context, pos):
        """
        Draw control point
        """
        prefs = context.user_preferences.addons["uv_magic_uv"].preferences
        cp_size = prefs.uvbb_cp_size
        offset = cp_size / 2
        verts = [
            [pos.x - offset, pos.y - offset],
            [pos.x - offset, pos.y + offset],
            [pos.x + offset, pos.y + offset],
            [pos.x + offset, pos.y - offset]
        ]
        bgl.glEnable(bgl.GL_BLEND)
        bgl.glBegin(bgl.GL_QUADS)
        bgl.glColor4f(1.0, 1.0, 1.0, 1.0)
        for (x, y) in verts:
            bgl.glVertex2f(x, y)
        bgl.glEnd()

    @staticmethod
    def draw_bb(_, context):
        """
        Draw bounding box
        """
        props = context.scene.muv_props.uvbb
        for cp in props.ctrl_points:
            MUV_UVBBRenderer.__draw_ctrl_point(
                context, mathutils.Vector(
                    context.region.view2d.view_to_region(cp.x, cp.y)))


class MUV_UVBBState(IntEnum):
    """
    Enum: State definition used by MUV_UVBBStateMgr
    """
    NONE = 0
    TRANSLATING = 1
    SCALING_1 = 2
    SCALING_2 = 3
    SCALING_3 = 4
    SCALING_4 = 5
    SCALING_5 = 6
    SCALING_6 = 7
    SCALING_7 = 8
    SCALING_8 = 9
    ROTATING = 10
    UNIFORM_SCALING_1 = 11
    UNIFORM_SCALING_2 = 12
    UNIFORM_SCALING_3 = 13
    UNIFORM_SCALING_4 = 14


class MUV_UVBBStateBase():
    """
    Custom class: Base class of state
    """

    def __init__(self):
        pass

    def update(self, context, event, ctrl_points, mouse_view):
        raise NotImplementedError


class MUV_UVBBStateNone(MUV_UVBBStateBase):
    """
    Custom class:
    No state
    Wait for event from mouse
    """

    def __init__(self, cmd_exec):
        super().__init__()
        self.__cmd_exec = cmd_exec

    def update(self, context, event, ctrl_points, mouse_view):
        """
        Update state
        """
        prefs = context.user_preferences.addons["uv_magic_uv"].preferences
        cp_react_size = prefs.uvbb_cp_react_size
        is_uscaling = context.scene.muv_uvbb_uniform_scaling
        if event.type == 'LEFTMOUSE':
            if event.value == 'PRESS':
                x, y = context.region.view2d.view_to_region(
                    mouse_view.x, mouse_view.y)
                for i, p in enumerate(ctrl_points):
                    px, py = context.region.view2d.view_to_region(p.x, p.y)
                    in_cp_x = (px + cp_react_size > x and
                               px - cp_react_size < x)
                    in_cp_y = (py + cp_react_size > y and
                               py - cp_react_size < y)
                    if in_cp_x and in_cp_y:
                        if is_uscaling:
                            arr = [1, 3, 6, 8]
                            if i in arr:
                                return (
                                    MUV_UVBBState.UNIFORM_SCALING_1 +
                                    arr.index(i)
                                )
                        else:
                            return MUV_UVBBState.TRANSLATING + i

        return MUV_UVBBState.NONE


class MUV_UVBBStateTranslating(MUV_UVBBStateBase):
    """
    Custom class: Translating state
    """

    def __init__(self, cmd_exec, ctrl_points):
        super().__init__()
        self.__cmd_exec = cmd_exec
        ix, iy = ctrl_points[0].x, ctrl_points[0].y
        self.__cmd_exec.append(MUV_UVBBTranslationCmd(ix, iy))

    def update(self, context, event, ctrl_points, mouse_view):
        if event.type == 'LEFTMOUSE':
            if event.value == 'RELEASE':
                return MUV_UVBBState.NONE
        if event.type == 'MOUSEMOVE':
            x, y = mouse_view.x, mouse_view.y
            self.__cmd_exec.top().set(x, y)
        return MUV_UVBBState.TRANSLATING


class MUV_UVBBStateScaling(MUV_UVBBStateBase):
    """
    Custom class: Scaling state
    """

    def __init__(self, cmd_exec, state, ctrl_points):
        super().__init__()
        self.__state = state
        self.__cmd_exec = cmd_exec
        dir_x_list = [1, 1, 1, 0, 0, 1, 1, 1]
        dir_y_list = [1, 0, 1, 1, 1, 1, 0, 1]
        idx = state - 2
        ix, iy = ctrl_points[idx + 1].x, ctrl_points[idx + 1].y
        ox, oy = ctrl_points[8 - idx].x, ctrl_points[8 - idx].y
        dir_x, dir_y = dir_x_list[idx], dir_y_list[idx]
        mat = self.__cmd_exec.execute(end=self.__cmd_exec.undo_size())
        self.__cmd_exec.append(
            MUV_UVBBScalingCmd(ix, iy, ox, oy, dir_x, dir_y, mat.inverted()))

    def update(self, context, event, ctrl_points, mouse_view):
        if event.type == 'LEFTMOUSE':
            if event.value == 'RELEASE':
                return MUV_UVBBState.NONE
        if event.type == 'MOUSEMOVE':
            x, y = mouse_view.x, mouse_view.y
            self.__cmd_exec.top().set(x, y)
        return self.__state


class MUV_UVBBStateUniformScaling(MUV_UVBBStateBase):
    """
    Custom class: Uniform Scaling state
    """

    def __init__(self, cmd_exec, state, ctrl_points):
        super().__init__()
        self.__state = state
        self.__cmd_exec = cmd_exec
        icp_idx = [1, 3, 6, 8]
        ocp_idx = [8, 6, 3, 1]
        idx = state - MUV_UVBBState.UNIFORM_SCALING_1
        ix, iy = ctrl_points[icp_idx[idx]].x, ctrl_points[icp_idx[idx]].y
        ox, oy = ctrl_points[ocp_idx[idx]].x, ctrl_points[ocp_idx[idx]].y
        mat = self.__cmd_exec.execute(end=self.__cmd_exec.undo_size())
        self.__cmd_exec.append(MUV_UVBBUniformScalingCmd(
            ix, iy, ox, oy, mat.inverted()))

    def update(self, context, event, ctrl_points, mouse_view):
        if event.type == 'LEFTMOUSE':
            if event.value == 'RELEASE':
                return MUV_UVBBState.NONE
        if event.type == 'MOUSEMOVE':
            x, y = mouse_view.x, mouse_view.y
            self.__cmd_exec.top().set(x, y)

        return self.__state


class MUV_UVBBStateRotating(MUV_UVBBStateBase):
    """
    Custom class: Rotating state
    """

    def __init__(self, cmd_exec, ctrl_points):
        super().__init__()
        self.__cmd_exec = cmd_exec
        ix, iy = ctrl_points[9].x, ctrl_points[9].y
        ox, oy = ctrl_points[0].x, ctrl_points[0].y
        self.__cmd_exec.append(MUV_UVBBRotationCmd(ix, iy, ox, oy))

    def update(self, context, event, ctrl_points, mouse_view):
        if event.type == 'LEFTMOUSE':
            if event.value == 'RELEASE':
                return MUV_UVBBState.NONE
        if event.type == 'MOUSEMOVE':
            x, y = mouse_view.x, mouse_view.y
            self.__cmd_exec.top().set(x, y)
        return MUV_UVBBState.ROTATING


class MUV_UVBBStateMgr():
    """
    Custom class: Manage state about this feature
    """

    def __init__(self, cmd_exec):
        self.__cmd_exec = cmd_exec          # command executer
        self.__state = MUV_UVBBState.NONE   # current state
        self.__state_obj = MUV_UVBBStateNone(self.__cmd_exec)

    def __update_state(self, next_state, ctrl_points):
        """
        Update state
        """

        if next_state == self.__state:
            return
        obj = None
        if next_state == MUV_UVBBState.TRANSLATING:
            obj = MUV_UVBBStateTranslating(self.__cmd_exec, ctrl_points)
        elif MUV_UVBBState.SCALING_1 <= next_state <= MUV_UVBBState.SCALING_8:
            obj = MUV_UVBBStateScaling(
                self.__cmd_exec, next_state, ctrl_points)
        elif next_state == MUV_UVBBState.ROTATING:
            obj = MUV_UVBBStateRotating(self.__cmd_exec, ctrl_points)
        elif next_state == MUV_UVBBState.NONE:
            obj = MUV_UVBBStateNone(self.__cmd_exec)
        elif (MUV_UVBBState.UNIFORM_SCALING_1 <= next_state <=
              MUV_UVBBState.UNIFORM_SCALING_4):
            obj = MUV_UVBBStateUniformScaling(
                self.__cmd_exec, next_state, ctrl_points)

        if obj is not None:
            self.__state_obj = obj

        self.__state = next_state

    def update(self, context, ctrl_points, event):
        mouse_region = mathutils.Vector((
            event.mouse_region_x, event.mouse_region_y))
        mouse_view = mathutils.Vector((context.region.view2d.region_to_view(
            mouse_region.x, mouse_region.y)))
        next_state = self.__state_obj.update(
            context, event, ctrl_points, mouse_view)
        self.__update_state(next_state, ctrl_points)


class MUV_UVBBUpdater(bpy.types.Operator):
    """
    Operation class: Update state and handle event by modal function
    """

    bl_idname = "uv.muv_uvbb_updater"
    bl_label = "UV Bounding Box Updater"
    bl_description = "Update UV Bounding Box"
    bl_options = {'REGISTER', 'UNDO'}

    def __init__(self):
        self.__timer = None
        self.__cmd_exec = MUV_UVBBCmdExecuter()         # Command executer
        self.__state_mgr = MUV_UVBBStateMgr(self.__cmd_exec)    # State Manager

    def __handle_add(self, context):
        if self.__timer is None:
            self.__timer = context.window_manager.event_timer_add(
                0.1, context.window)
            context.window_manager.modal_handler_add(self)
        MUV_UVBBRenderer.handle_add(self, context)

    def __handle_remove(self, context):
        MUV_UVBBRenderer.handle_remove()
        if self.__timer is not None:
            context.window_manager.event_timer_remove(self.__timer)
            self.__timer = None

    def __get_uv_info(self, context):
        """
        Get UV coordinate
        """
        obj = context.active_object
        uv_info = []
        bm = bmesh.from_edit_mesh(obj.data)
        if muv_common.check_version(2, 73, 0) >= 0:
            bm.faces.ensure_lookup_table()
        if not bm.loops.layers.uv:
            return None
        uv_layer = bm.loops.layers.uv.verify()
        for f in bm.faces:
            if f.select:
                for i, l in enumerate(f.loops):
                    uv_info.append((f.index, i, l[uv_layer].uv.copy()))
        if len(uv_info) == 0:
            return None
        return uv_info

    def __get_ctrl_point(self, uv_info_ini):
        """
        Get control point
        """
        left = MAX_VALUE
        right = -MAX_VALUE
        top = -MAX_VALUE
        bottom = MAX_VALUE

        for info in uv_info_ini:
            uv = info[2]
            if uv.x < left:
                left = uv.x
            if uv.x > right:
                right = uv.x
            if uv.y < bottom:
                bottom = uv.y
            if uv.y > top:
                top = uv.y

        points = [
            mathutils.Vector((
                (left + right) * 0.5, (top + bottom) * 0.5, 0.0
            )),
            mathutils.Vector((left, top, 0.0)),
            mathutils.Vector((left, (top + bottom) * 0.5, 0.0)),
            mathutils.Vector((left, bottom, 0.0)),
            mathutils.Vector(((left + right) * 0.5, top, 0.0)),
            mathutils.Vector(((left + right) * 0.5, bottom, 0.0)),
            mathutils.Vector((right, top, 0.0)),
            mathutils.Vector((right, (top + bottom) * 0.5, 0.0)),
            mathutils.Vector((right, bottom, 0.0)),
            mathutils.Vector(((left + right) * 0.5, top + 0.03, 0.0))
        ]

        return points

    def __update_uvs(self, context, uv_info_ini, trans_mat):
        """
        Update UV coordinate
        """
        obj = context.active_object
        bm = bmesh.from_edit_mesh(obj.data)
        if muv_common.check_version(2, 73, 0) >= 0:
            bm.faces.ensure_lookup_table()
        if not bm.loops.layers.uv:
            return
        uv_layer = bm.loops.layers.uv.verify()
        for info in uv_info_ini:
            fidx = info[0]
            lidx = info[1]
            uv = info[2]
            v = mathutils.Vector((uv.x, uv.y, 0.0))
            av = trans_mat * v
            bm.faces[fidx].loops[lidx][uv_layer].uv = mathutils.Vector(
                (av.x, av.y))

    def __update_ctrl_point(self, ctrl_points_ini, trans_mat):
        """
        Update control point
        """
        return [trans_mat * cp for cp in ctrl_points_ini]

    def modal(self, context, event):
        props = context.scene.muv_props.uvbb
        muv_common.redraw_all_areas()
        if props.running is False:
            self.__handle_remove(context)
            return {'FINISHED'}
        if event.type == 'TIMER':
            trans_mat = self.__cmd_exec.execute()
            self.__update_uvs(context, props.uv_info_ini, trans_mat)
            props.ctrl_points = self.__update_ctrl_point(
                props.ctrl_points_ini, trans_mat)

        self.__state_mgr.update(context, props.ctrl_points, event)

        return {'PASS_THROUGH'}

    def execute(self, context):
        props = context.scene.muv_props.uvbb

        if props.running is True:
            props.running = False
            return {'FINISHED'}

        props.uv_info_ini = self.__get_uv_info(context)
        if props.uv_info_ini is None:
            return {'CANCELLED'}
        props.ctrl_points_ini = self.__get_ctrl_point(props.uv_info_ini)
        trans_mat = self.__cmd_exec.execute()
        # Update is needed in order to display control point
        self.__update_uvs(context, props.uv_info_ini, trans_mat)
        props.ctrl_points = self.__update_ctrl_point(
            props.ctrl_points_ini, trans_mat)
        self.__handle_add(context)
        props.running = True

        return {'RUNNING_MODAL'}


class IMAGE_PT_MUV_UVBB(bpy.types.Panel):
    """
    Panel class: UV Bounding Box Menu on Property Panel on UV/ImageEditor
    """

    bl_space_type = 'IMAGE_EDITOR'
    bl_region_type = 'UI'
    bl_label = "UV Bounding Box"
    bl_context = 'mesh_edit'

    @classmethod
    def poll(cls, context):
        prefs = context.user_preferences.addons["uv_magic_uv"].preferences
        return prefs.enable_uvbb

    def draw_header(self, _):
        layout = self.layout
        layout.label(text="", icon='IMAGE_COL')

    def draw(self, context):
        sc = context.scene
        props = sc.muv_props.uvbb
        layout = self.layout
        if props.running is False:
            layout.operator(
                MUV_UVBBUpdater.bl_idname, text="Display UV Bounding Box",
                icon='PLAY')
        else:
            layout.operator(
                MUV_UVBBUpdater.bl_idname, text="Hide UV Bounding Box",
                icon='PAUSE')
        layout.prop(sc, "muv_uvbb_uniform_scaling", text="Uniform Scaling")



# ##### BEGIN GPL LICENSE BLOCK #####
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# Note: the Operator LoadRandomEditOperator was removed since is not
# working. If it is fixed, reimplemented it can be reintroduced later

import bpy
from bpy.types import (
        Operator,
        Panel,
        )
from . import functions


# classes
class RandomScratchOperator(Operator):
    bl_idname = "sequencer.randomscratchoperator"
    bl_label = "Random Scratch Operator"
    bl_description = "Random Scratch Operator"

    @classmethod
    def poll(self, context):
        strip = functions.act_strip(context)
        scn = context.scene
        if scn and scn.sequence_editor and scn.sequence_editor.active_strip:
            return strip.type in ('META')
        else:
            return False

    def invoke(self, context, event):
        preferences = context.user_preferences
        random_frames = preferences.addons[__package__].preferences.random_frames

        sce = context.scene
        seq = sce.sequence_editor
        markers = sce.timeline_markers

        if seq:
            strip = seq.active_strip
            if strip is not None:
                if "IN" and "OUT" in markers:
                    sin = markers["IN"].frame
                    sout = markers["OUT"].frame

                    # select active strip
                    strip = context.scene.sequence_editor.active_strip
                    stripname = strip.name
                    # collect strip names inside the meta
                    stripnames = []
                    stripnames.append(strip.name)
                    for i in seq.active_strip.sequences:
                        stripnames.append(i.name)
                    # get strip channel
                    channel = strip.channel
                    repeat = range(int((sout - sin) / random_frames))
                    print(sin, sout, sout - sin, (sout - sin) / random_frames, repeat)

                    for i in repeat:
                        # select all related strips
                        for j in stripnames:
                            strip = seq.sequences_all[j]
                            strip.select = True
                        strip = seq.sequences_all[stripname]
                        seq.active_strip = strip
                        # deselect all other strips
                        for j in context.selected_editable_sequences:
                            if j.name not in stripnames:
                                j.select = False
                        a = bpy.ops.sequencer.duplicate_move()
                        # select new strip
                        newstrip = seq.active_strip
                        # deselect all other strips

                        for j in context.selected_editable_sequences:
                            if j.name != newstrip.name:
                                j.select = False
                        # random cut
                        newstrip.frame_start = sin + i * random_frames
                        rand = functions.randomframe(newstrip)
                        functions.triminout(newstrip, rand, rand + random_frames)
                        newstrip.frame_start = i * random_frames + sin - newstrip.frame_offset_start
                        newstrip.channel = channel + 1
                else:
                    self.report({'WARNING'}, "There is no IN and OUT Markers")
            bpy.ops.sequencer.reload()

        return {'FINISHED'}


class RandomEditorPanel(Panel):
    bl_label = "Random Editor"
    bl_idname = "OBJECT_PT_RandomEditor"
    bl_space_type = 'SEQUENCE_EDITOR'
    bl_region_type = 'UI'

    @classmethod
    def poll(self, context):
        if context.space_data.view_type in {'SEQUENCER',
                                            'SEQUENCER_PREVIEW'}:
            strip = functions.act_strip(context)
            scn = context.scene
            preferences = context.user_preferences
            prefs = preferences.addons[__package__].preferences
            if scn and scn.sequence_editor and scn.sequence_editor.active_strip:
                if prefs.use_random_editor:
                    return strip.type in ('META')
        else:
            return False

    def draw_header(self, context):
        layout = self.layout
        layout.label(text="", icon="MOD_BUILD")

    def draw(self, context):

        preferences = context.user_preferences
        prefs = preferences.addons[__package__].preferences

        layout = self.layout
        col = layout.column(align=True)
        col.label("Cut duration:")
        col.prop(prefs, "random_frames")
        col.operator("sequencer.randomscratchoperator")
